"""Mapping dictionaries from hydromt to hydromt_wflow to Wflow.jl names."""

# Names that cannot be read from TOML but that HydroMT needs for model building
# {hydromt_name: staticmap_name}
HYDROMT_NAMES_DEFAULT: dict[str, str] = {
    "glacareas": "meta_glacier_area_id",
}

HYDROMT_NAMES_DEFAULT_SEDIMENT: dict[str, str] = {
    "elevtn": "land_elevation",
}

HYDROMT_NAMES_COMMON: dict[str, str] = {
    "subelv": "meta_subgrid_elevation",
    "uparea": "meta_upstream_area",
    "subare": "meta_subgrid_area",
    "strord": "meta_streamorder",
    "x_out": "meta_subgrid_outlet_x",
    "y_out": "meta_subgrid_outlet_y",
    "landuse": "meta_landuse",
    "soil_texture": "meta_soil_texture",
}

HYDROMT_NAMES_DEFAULT.update(HYDROMT_NAMES_COMMON)
HYDROMT_NAMES_DEFAULT_SEDIMENT.update(HYDROMT_NAMES_COMMON)

# Variables in the input section rather than input.static
WFLOW_VARS_IN_INPUT = [
    "basin__local_drain_direction",
    "subbasin_location__count",
    "river_location__mask",
    "reservoir_area__count",
    "reservoir_location__count",
    "reservoir_lower_location__count",
]

# Link between staticmap names, hydromt name (if any)
# and Wflow.jl variables for v0x and v1x (if not present, None)
# {staticmap_name: (wflow_v0, wflow_v1, hydromt_name)} or
# {staticmap_name: (wflow_v0, wflow_v1)}
WFLOW_NAMES: dict[str, dict[str, str | None]] = {
    # general input
    "subcatchment": {
        "wflow_v0": "subcatchment",
        "wflow_v1": "subbasin_location__count",
        "hydromt_name": "basins",
    },
    "local_drain_direction": {
        "wflow_v0": "ldd",
        "wflow_v1": "basin__local_drain_direction",
        "hydromt_name": "flwdir",
    },
    "lake_area_id": {
        "wflow_v0": "lateral.river.lake.areas",
        "wflow_v1": None,
        "hydromt_name": "lake_area_id",
    },
    "lake_outlet_id": {
        "wflow_v0": "lateral.river.lake.locs",
        "wflow_v1": None,
        "hydromt_name": "lake_outlet_id",
    },
    "reservoir_area_id": {
        "wflow_v0": "lateral.river.reservoir.areas",
        "wflow_v1": "reservoir_area__count",
        "hydromt_name": "reservoir_area_id",
    },
    "reservoir_outlet_id": {
        "wflow_v0": "lateral.river.reservoir.locs",
        "wflow_v1": "reservoir_location__count",
        "hydromt_name": "reservoir_outlet_id",
    },
    "river_mask": {
        "wflow_v0": "river_location",
        "wflow_v1": "river_location__mask",
        "hydromt_name": "rivmsk",
    },
    "pits": {
        "wflow_v0": "pits",
        "wflow_v1": "basin_pit_location__mask ",
    },
    # atmosphere / forcing
    "precip": {
        "wflow_v0": "vertical.precipitation",
        "wflow_v1": "atmosphere_water__precipitation_volume_flux",
        "hydromt_name": "precip",
    },
    "temp": {
        "wflow_v0": "vertical.temperature",
        "wflow_v1": "atmosphere_air__temperature",
        "hydromt_name": "temp",
    },
    "pet": {
        "wflow_v0": "vertical.potential_evaporation",
        "wflow_v1": "land_surface_water__potential_evaporation_volume_flux",
        "hydromt_name": "pet",
    },
    # snow
    "snow_degree_day_coefficient": {
        "wflow_v0": "vertical.cfmax",
        "wflow_v1": "snowpack__degree_day_coefficient",
    },
    "snow_tt": {
        "wflow_v0": "vertical.tt",
        "wflow_v1": "atmosphere_air__snowfall_temperature_threshold",
    },
    "snow_tti": {
        "wflow_v0": "vertical.tti",
        "wflow_v1": "atmosphere_air__snowfall_temperature_interval",
    },
    "snow_ttm": {
        "wflow_v0": "vertical.ttm",
        "wflow_v1": "snowpack__melting_temperature_threshold",
    },
    "snow_whc": {
        "wflow_v0": "vertical.water_holding_capacity",
        "wflow_v1": "snowpack__liquid_water_holding_capacity",
    },
    # glacier
    "glacier_fraction": {
        "wflow_v0": "vertical.glacierfrac",
        "wflow_v1": "glacier_surface__area_fraction",
        "hydromt_name": "glacfracs",
    },
    "glacier_initial_leq_depth": {
        "wflow_v0": "vertical.glacierstore",
        "wflow_v1": "glacier_ice__initial_leq_depth",
        "hydromt_name": "glacstore",
    },
    "glacier_ttm": {
        "wflow_v0": ["vertical.g_ttm", "vertical.g_tt"],
        "wflow_v1": "glacier_ice__melting_temperature_threshold",
    },
    "glacier_degree_day_coefficient": {
        "wflow_v0": "vertical.g_cfmax",
        "wflow_v1": "glacier_ice__degree_day_coefficient",
    },
    "glacier_snow_to_ice_fraction": {
        "wflow_v0": "vertical.g_sifrac",
        "wflow_v1": "glacier_firn_accumulation__snowpack_dry_snow_leq_depth_fraction",
    },
    # vegetation
    "vegetation_gash_e_r": {
        "wflow_v0": "vertical.e_r",
        "wflow_v1": "vegetation_canopy_water__mean_evaporation_to_mean_precipitation_ratio",  # noqa: E501
    },
    "vegetation_kext": {
        "wflow_v0": "vertical.kext",
        "wflow_v1": "vegetation_canopy__light_extinction_coefficient",
    },
    "vegetation_leaf_area_index": {
        "wflow_v0": "vertical.leaf_area_index",
        "wflow_v1": "vegetation__leaf_area_index",
        "hydromt_name": "LAI",
    },
    "vegetation_root_depth": {
        "wflow_v0": "vertical.rootingdepth",
        "wflow_v1": "vegetation_root__depth",
    },
    "vegetation_leaf_storage": {
        "wflow_v0": "vertical.specific_leaf",
        "wflow_v1": "vegetation__specific_leaf_storage",
        "hydromt_name": "leaf_storage",
    },
    "vegetation_wood_storage": {
        "wflow_v0": "vertical.storage_wood",
        "wflow_v1": "vegetation_wood_water__storage_capacity",
        "hydromt_name": "wood_storage",
    },
    "vegetation_crop_factor": {
        "wflow_v0": ["vertical.kc", "vertical.etreftopot"],
        "wflow_v1": "vegetation__crop_factor",
    },
    "vegetation_feddes_alpha_h1": {
        "wflow_v0": "vertical.alpha_h1",
        "wflow_v1": "vegetation_root__feddes_critical_pressure_head_h1_reduction_coefficient",  # noqa: E501
    },
    "vegetation_feddes_h1": {
        "wflow_v0": "vertical.h1",
        "wflow_v1": "vegetation_root__feddes_critical_pressure_head_h1",
    },
    "vegetation_feddes_h2": {
        "wflow_v0": "vertical.h2",
        "wflow_v1": "vegetation_root__feddes_critical_pressure_head_h2",
    },
    "vegetation_feddes_h3_high": {
        "wflow_v0": "vertical.h3_high",
        "wflow_v1": "vegetation_root__feddes_critical_pressure_head_h3_high",
    },
    "vegetation_feddes_h3_low": {
        "wflow_v0": "vertical.h3_low",
        "wflow_v1": "vegetation_root__feddes_critical_pressure_head_h3_low",
    },
    "vegetation_feddes_h4": {
        "wflow_v0": "vertical.h4",
        "wflow_v1": "vegetation_root__feddes_critical_pressure_head_h4",
    },
    # soil
    "soil_brooks_corey_c": {
        "wflow_v0": "vertical.c",
        "wflow_v1": "soil_layer_water__brooks_corey_exponent",
        "hydromt_name": "soil_brooks_corey_c",
    },
    "soil_cf": {
        "wflow_v0": "vertical.cf_soil",
        "wflow_v1": "soil_surface_water__infiltration_reduction_parameter",
    },
    "soil_ksat_vertical": {
        "wflow_v0": ["vertical.kv_0", "vertical.kv₀"],
        "wflow_v1": "soil_surface_water__vertical_saturated_hydraulic_conductivity",
        "hydromt_name": "ksat_vertical",
    },
    "soil_ksat_vertical_factor": {
        "wflow_v0": "vertical.kvfrac",
        "wflow_v1": "soil_water__vertical_saturated_hydraulic_conductivity_factor",
    },
    "subsurface_ksat_horizontal_ratio": {
        "wflow_v0": "lateral.subsurface.ksathorfrac",
        "wflow_v1": "subsurface_water__horizontal_to_vertical_saturated_hydraulic_conductivity_ratio",  # noqa: E501
    },
    "soil_f": {
        "wflow_v0": "vertical.f",
        "wflow_v1": "soil_water__vertical_saturated_hydraulic_conductivity_scale_parameter",  # noqa: E501
        "hydromt_name": "f",
    },
    "soil_compacted_infiltration_capacity": {
        "wflow_v0": "vertical.infiltcappath",
        "wflow_v1": "compacted_soil_surface_water__infiltration_capacity",
    },
    "InfiltCapSoil": {
        "wflow_v0": "vertical.infiltcapsoil",
        "wflow_v1": None,
    },
    "soil_theta_r": {
        "wflow_v0": ["vertical.theta_r", "vertical.θᵣ"],
        "wflow_v1": "soil_water__residual_volume_fraction",
        "hydromt_name": "theta_r",
    },
    "soil_theta_s": {
        "wflow_v0": ["vertical.theta_s", "vertical.θₛ"],
        "wflow_v1": "soil_water__saturated_volume_fraction",
        "hydromt_name": "theta_s",
    },
    "soil_max_leakage": {
        "wflow_v0": "vertical.maxleakage",
        "wflow_v1": "soil_water_saturated_zone_bottom__max_leakage_volume_flux",
    },
    "soil_compacted_fraction": {
        "wflow_v0": "vertical.pathfrac",
        "wflow_v1": "compacted_soil__area_fraction",
    },
    "soil_rootdistpar": {
        "wflow_v0": "vertical.rootdistpar",
        "wflow_v1": "soil_wet_root__sigmoid_function_shape_parameter",
    },
    "soil_thickness": {
        "wflow_v0": "vertical.soilthickness",
        "wflow_v1": "soil__thickness",
    },
    # land
    "land_water_fraction": {
        "wflow_v0": "vertical.waterfrac",
        "wflow_v1": "land_water_covered__area_fraction",
    },
    "demand_allocation_area_id": {
        "wflow_v0": "vertical.allocation.areas",
        "wflow_v1": "land_water_allocation_area__count",
        "hydromt_name": "allocation_areas",
    },
    "demand_surface_water_ratio": {
        "wflow_v0": "vertical.allocation.frac_sw_used",
        "wflow_v1": "land_surface_water__withdrawal_fraction",
    },
    "demand_domestic_gross": {
        "wflow_v0": "vertical.domestic.demand_gross",
        "wflow_v1": "domestic__gross_water_demand_volume_flux",
        "hydromt_name": "domestic_gross",
    },
    "demand_domestic_net": {
        "wflow_v0": "vertical.domestic.demand_net",
        "wflow_v1": "domestic__net_water_demand_volume_flux",
        "hydromt_name": "domestic_net",
    },
    "demand_industry_gross": {
        "wflow_v0": "vertical.industry.demand_gross",
        "wflow_v1": "industry__gross_water_demand_volume_flux",
        "hydromt_name": "industry_gross",
    },
    "demand_industry_net": {
        "wflow_v0": "vertical.industry.demand_net",
        "wflow_v1": "industry__net_water_demand_volume_flux",
        "hydromt_name": "industry_net",
    },
    "demand_livestock_gross": {
        "wflow_v0": "vertical.livestock.demand_gross",
        "wflow_v1": "livestock__gross_water_demand_volume_flux",
        "hydromt_name": "livestock_gross",
    },
    "demand_livestock_net": {
        "wflow_v0": "vertical.livestock.demand_net",
        "wflow_v1": "livestock__net_water_demand_volume_flux",
        "hydromt_name": "livestock_net",
    },
    "demand_paddy_h_min": {
        "wflow_v0": "vertical.paddy.h_min",
        "wflow_v1": "irrigated_paddy__min_depth",
    },
    "demand_paddy_h_opt": {
        "wflow_v0": "vertical.paddy.h_opt",
        "wflow_v1": "irrigated_paddy__optimal_depth",
    },
    "demand_paddy_h_max": {
        "wflow_v0": "vertical.paddy.h_max",
        "wflow_v1": "irrigated_paddy__max_depth",
    },
    "demand_paddy_irrigated_mask": {
        "wflow_v0": "vertical.paddy.irrigation_areas",
        "wflow_v1": "irrigated_paddy_area__count",
    },
    "demand_paddy_irrigation_trigger": {
        "wflow_v0": "vertical.paddy.irrigation_trigger",
        "wflow_v1": "irrigated_paddy__irrigation_trigger_flag",
    },
    "demand_nonpaddy_irrigated_mask": {
        "wflow_v0": "vertical.nonpaddy.irrigation_areas",
        "wflow_v1": "irrigated_non_paddy_area__count",
    },
    "demand_nonpaddy_irrigation_trigger": {
        "wflow_v0": "vertical.nonpaddy.irrigation_trigger",
        "wflow_v1": "irrigated_non_paddy__irrigation_trigger_flag",
    },
    # land surface water flow
    "land_manning_n": {
        "wflow_v0": "lateral.land.n",
        "wflow_v1": "land_surface_water_flow__manning_n_parameter",
    },
    "land_elevation": {
        "wflow_v0": "lateral.land.elevation",
        "wflow_v1": "land_surface_water_flow__ground_elevation",
        "hydromt_name": "elevtn",
    },
    "land_slope": {
        "wflow_v0": "lateral.land.slope",
        "wflow_v1": "land_surface__slope",
        "hydromt_name": "lndslp",
    },
    # river
    "floodplain_volume": {
        "wflow_v0": "lateral.river.floodplain.volume",
        "wflow_v1": "floodplain_water__sum_of_volume_per_depth",
        "hydromt_name": "floodplain_volume",
    },
    "floodplain_manning_n": {
        "wflow_v0": "lateral.river.floodplain.n",
        "wflow_v1": "floodplain_water_flow__manning_n_parameter",
    },
    "river_bank_elevation": {
        "wflow_v0": "lateral.river.bankfull_elevation",
        "wflow_v1": "river_bank_water__elevation",
        "hydromt_name": "hydrodem",
    },
    "river_inflow": {
        "wflow_v0": "lateral.river.inflow",
        "wflow_v1": "river_water__external_inflow_volume_flow_rate",
    },
    "river_depth": {
        "wflow_v0": "lateral.river.bankfull_depth",
        "wflow_v1": "river_bank_water__depth",
        "hydromt_name": "rivdph",
    },
    "river_length": {
        "wflow_v0": "lateral.river.length",
        "wflow_v1": "river__length",
        "hydromt_name": "rivlen",
    },
    "river_manning_n": {
        "wflow_v0": "lateral.river.n",
        "wflow_v1": "river_water_flow__manning_n_parameter",
    },
    "river_slope": {
        "wflow_v0": "lateral.river.slope",
        "wflow_v1": "river__slope",
        "hydromt_name": "rivslp",
    },
    "river_width": {
        "wflow_v0": "lateral.river.width",
        "wflow_v1": "river__width",
        "hydromt_name": "rivwth",
    },
    # lakes
    "lake_area": {
        "wflow_v0": "lateral.river.lake.area",
        "wflow_v1": None,
        "hydromt_name": "lake_area",
    },
    "reservoir_initial_depth": {
        "wflow_v0": "lateral.river.lake.waterlevel",
        "wflow_v1": "reservoir_water_surface__initial_elevation",
        "hydromt_name": "reservoir_initial_depth",
    },
    "reservoir_outflow_threshold": {
        "wflow_v0": "lateral.river.lake.threshold",
        "wflow_v1": "reservoir_water_flow_threshold_level__elevation",
        "hydromt_name": "reservoir_outflow_threshold",
    },
    "reservoir_b": {
        "wflow_v0": "lateral.river.lake.b",
        "wflow_v1": "reservoir_water__rating_curve_coefficient",
        "hydromt_name": "reservoir_b",
    },
    "reservoir_e": {
        "wflow_v0": "lateral.river.lake.e",
        "wflow_v1": "reservoir_water__rating_curve_exponent",
        "hydromt_name": "reservoir_e",
    },
    "reservoir_rating_curve": {
        "wflow_v0": "lateral.river.lake.outflowfunc",
        "wflow_v1": "reservoir_water__rating_curve_type_count",
        "hydromt_name": "reservoir_rating_curve",
    },
    "reservoir_storage_curve": {
        "wflow_v0": "lateral.river.lake.storfunc",
        "wflow_v1": "reservoir_water__storage_curve_type_count",
        "hydromt_name": "reservoir_storage_curve",
    },
    "reservoir_lower_id": {
        "wflow_v0": "lateral.river.lake.linkedlakelocs",
        "wflow_v1": "reservoir_lower_location__count",
        "hydromt_name": "reservoir_lower_id",
    },
    # reservoirs
    "reservoir_area": {
        "wflow_v0": "lateral.river.reservoir.area",
        "wflow_v1": "reservoir_surface__area",
        "hydromt_name": "reservoir_area",
    },
    "reservoir_demand": {
        "wflow_v0": "lateral.river.reservoir.demand",
        "wflow_v1": "reservoir_water_demand__required_downstream_volume_flow_rate",
        "hydromt_name": "reservoir_demand",
    },
    "reservoir_max_release": {
        "wflow_v0": "lateral.river.reservoir.maxrelease",
        "wflow_v1": "reservoir_water_release_below_spillway__max_volume_flow_rate",
        "hydromt_name": "reservoir_max_release",
    },
    "reservoir_max_volume": {
        "wflow_v0": "lateral.river.reservoir.maxvolume",
        "wflow_v1": "reservoir_water__max_volume",
        "hydromt_name": "reservoir_max_volume",
    },
    "reservoir_target_full_fraction": {
        "wflow_v0": "lateral.river.reservoir.targetfullfrac",
        "wflow_v1": "reservoir_water__target_full_volume_fraction",
        "hydromt_name": "reservoir_target_full_fraction",
    },
    "reservoir_target_min_fraction": {
        "wflow_v0": "lateral.river.reservoir.targetminfrac",
        "wflow_v1": "reservoir_water__target_min_volume_fraction",
        "hydromt_name": "reservoir_target_min_fraction",
    },
    # gwf
    "altitude": {
        "wflow_v0": "altitude",
        "wflow_v1": "land_surface__elevation",
    },
    "groundwater_constant_head": {
        "wflow_v0": "lateral.subsurface.constant_head",
        "wflow_v1": "model_constant_boundary_condition__hydraulic_head",
    },
    "groundwater_ksat_horizontal": {
        "wflow_v0": "lateral.subsurface.conductivity",
        "wflow_v1": "subsurface_surface_water__horizontal_saturated_hydraulic_conductivity",  # noqa: E501
    },
    "river_bed_elevation": {
        "wflow_v0": "lateral.subsurface.river_bottom",
        "wflow_v1": "river_bottom__elevation",
    },
    "river_bed_exfiltration_conductance": {
        "wflow_v0": "lateral.subsurface.infiltration_conductance",
        "wflow_v1": "river_water__infiltration_conductance",
    },
    "river_bed_infiltration_conductance": {
        "wflow_v0": "lateral.subsurface.exfiltration_conductance",
        "wflow_v1": "river_water__exfiltration_conductance",
    },
    "groundwater_specific_yield": {
        "wflow_v0": "lateral.subsurface.specific_yield",
        "wflow_v1": "subsurface_water__specific_yield",
    },
    "groundwater_f": {
        "wflow_v0": "lateral.subsurface.gwf_f",
        "wflow_v1": "subsurface__horizontal_saturated_hydraulic_conductivity_scale_parameter",  # noqa: E501
    },
    "drain": {
        "wflow_v0": "lateral.subsurface.drain",
        "wflow_v1": "land_drain_location__mask",
    },
    "drain_conductance": {
        "wflow_v0": "lateral.subsurface.drain_conductance",
        "wflow_v1": "land_drain__conductance",
    },
    "drain_elevation": {
        "wflow_v0": "lateral.subsurface.drain_elevation",
        "wflow_v1": "land_drain__elevation",
    },
}

WFLOW_STATES_NAMES: dict[str, dict[str, str | None]] = {
    "vegetation_water_depth": {
        "wflow_v0": "vertical.canopystorage",
        "wflow_v1": "vegetation_canopy_water__depth",
    },
    "soil_saturated_depth": {
        "wflow_v0": "vertical.satwaterdepth",
        "wflow_v1": "soil_water_saturated_zone__depth",
    },
    "soil_unsaturated_depth": {
        "wflow_v0": "vertical.ustorelayerdepth",
        "wflow_v1": "soil_layer_water_unsaturated_zone__depth",
    },
    "soil_temp": {
        "wflow_v0": "vertical.tsoil",
        "wflow_v1": "soil_surface__temperature",
    },
    "snow_leq_depth": {
        "wflow_v0": "vertical.snow",
        "wflow_v1": "snowpack_dry_snow__leq_depth",
    },
    "snow_water_depth": {
        "wflow_v0": "vertical.snowwater",
        "wflow_v1": "snowpack_liquid_water__depth",
    },
    "glacier_leq_depth": {
        "wflow_v0": "vertical.glacierstore",
        "wflow_v1": "glacier_ice__leq_depth",
    },
    "land_instantaneous_q": {
        "wflow_v0": "lateral.land.q",
        "wflow_v1": "land_surface_water__instantaneous_volume_flow_rate",
    },
    "land_instantaneous_qx": {
        "wflow_v0": "lateral.land.qx",
        "wflow_v1": "land_surface_water__x_component_of_instantaneous_volume_flow_rate",
    },
    "land_instantaneous_qy": {
        "wflow_v0": "lateral.land.qy",
        "wflow_v1": "land_surface_water__y_component_of_instantaneous_volume_flow_rate",
    },
    "land_h": {
        "wflow_v0": "lateral.land.h",
        "wflow_v1": "land_surface_water__depth",
    },
    "h_av_land": {
        "wflow_v0": "lateral.land.h_av",
        "wflow_v1": None,
    },
    "groundwater_head": {
        "wflow_v0": "lateral.subsurface.flow.aquifer.head",
        "wflow_v1": "subsurface_water__hydraulic_head",
    },
    "subsurface_q": {
        "wflow_v0": "lateral.subsurface.ssf",
        "wflow_v1": "subsurface_water__volume_flow_rate",
    },
    "river_instantaneous_q": {
        "wflow_v0": "lateral.river.q",
        "wflow_v1": "river_water__instantaneous_volume_flow_rate",
    },
    "river_h": {
        "wflow_v0": "lateral.river.h",
        "wflow_v1": "river_water__depth",
    },
    "h_av_river": {
        "wflow_v0": "lateral.river.h_av",
        "wflow_v1": None,
    },
    "floodplain_instantaneous_q": {
        "wflow_v0": "lateral.river.floodplain.q",
        "wflow_v1": "floodplain_water__instantaneous_volume_flow_rate",
    },
    "floodplain_h": {
        "wflow_v0": "lateral.river.floodplain.h",
        "wflow_v1": "floodplain_water__depth",
    },
    "lake_water_level": {
        "wflow_v0": "lateral.river.lake.waterlevel",
        "wflow_v1": "reservoir_water_surface__elevation",
    },
    "reservoir_volume": {
        "wflow_v0": "lateral.river.reservoir.volume",
        "wflow_v1": None,
    },
    "demand_paddy_h": {
        "wflow_v0": "vertical.paddy.h",
        "wflow_v1": "paddy_surface_water__depth",
    },
}

WFLOW_SEDIMENT_NAMES: dict[str, dict[str, str | None]] = {
    # general input
    "subcatchment": {
        "wflow_v0": "subcatchment",
        "wflow_v1": "subbasin_location__count",
        "hydromt_name": "basins",
    },
    "local_drain_direction": {
        "wflow_v0": "ldd",
        "wflow_v1": "basin__local_drain_direction",
        "hydromt_name": "flwdir",
    },
    "lake_area_id": {
        "wflow_v0": "vertical.lakeareas",
        "wflow_v1": None,
        "hydromt_name": "lake_area_id",
    },
    "lake_outlet_id": {
        "wflow_v0": "lateral.river.lakelocs",
        "wflow_v1": None,
        "hydromt_name": "lake_outlet_id",
    },
    "reservoir_area_id": {
        "wflow_v0": "vertical.resareas",
        "wflow_v1": "reservoir_area__count",
        "hydromt_name": "reservoir_area_id",
    },
    "reservoir_outlet_id": {
        "wflow_v0": "lateral.river.reslocs",
        "wflow_v1": "reservoir_location__count",
        "hydromt_name": "reservoir_outlet_id",
    },
    "river_mask": {
        "wflow_v0": "river_location",
        "wflow_v1": "river_location__mask",
        "hydromt_name": "rivmsk",
    },
    # forcing
    "precip": {
        "wflow_v0": "vertical.precipitation",
        "wflow_v1": "atmosphere_water__precipitation_volume_flux",
        "hydromt_name": "precip",
    },
    "vegetation_interception": {
        "wflow_v0": "vertical.interception",
        "wflow_v1": "vegetation_canopy_water__interception_volume_flux",
        "hydromt_name": "interception",
    },
    "land_h": {
        "wflow_v0": "vertical.h_land",
        "wflow_v1": "land_surface_water__depth",
    },
    "land_q": {
        "wflow_v0": "vertical.q_land",
        "wflow_v1": "land_surface_water__volume_flow_rate",
    },
    "river_h": {
        "wflow_v0": "lateral.river.h_riv",
        "wflow_v1": "river_water__depth",
    },
    "river_q": {
        "wflow_v0": "lateral.river.q_riv",
        "wflow_v1": "river_water__volume_flow_rate",
    },
    # land properties
    "land_slope": {
        "wflow_v0": "lateral.land.slope",
        "wflow_v1": "land_surface__slope",
        "hydromt_name": "lndslp",
    },
    # river properties
    "river_length": {
        "wflow_v0": "lateral.river.length",
        "wflow_v1": "river__length",
        "hydromt_name": "rivlen",
    },
    "river_slope": {
        "wflow_v0": "lateral.river.slope",
        "wflow_v1": "river__slope",
        "hydromt_name": "rivslp",
    },
    "river_width": {
        "wflow_v0": "lateral.river.width",
        "wflow_v1": "river__width",
        "hydromt_name": "rivwth",
    },
    # reservoirs
    "reservoir_area": {
        "wflow_v0": "lateral.river.resarea",
        "wflow_v1": "reservoir_surface__area",
        "hydromt_name": "reservoir_area",
    },
    "lake_area": {
        "wflow_v0": "lateral.river.lakearea",
        "wflow_v1": None,
        "hydromt_name": "lake_area",
    },
    "reservoir_trapping_efficiency": {
        "wflow_v0": "lateral.river.restrapeff",
        "wflow_v1": "reservoir_water_sediment__bedload_trapping_efficiency",
        "hydromt_name": "reservoir_trapping_efficiency",
    },
    # soil erosion
    "vegetation_gap_fraction": {
        "wflow_v0": "vertical.canopygapfraction",
        "wflow_v1": "vegetation_canopy__gap_fraction",
    },
    "vegetation_height": {
        "wflow_v0": "vertical.canopyheight",
        "wflow_v1": "vegetation_canopy__height",
    },
    "vegetation_kext": {
        "wflow_v0": "vertical.kext",
        "wflow_v1": None,
    },
    "leaf_storage": {
        "wflow_v0": "vertical.specific_leaf",
        "wflow_v1": None,
        "hydromt_name": "leaf_storage",
    },
    "vegetation_wood_storage": {
        "wflow_v0": "vertical.storage_wood",
        "wflow_v1": None,
        "hydromt_name": "wood_storage",
    },
    "soil_compacted_fraction": {
        "wflow_v0": "vertical.pathfrac",
        "wflow_v1": "compacted_soil__area_fraction",
    },
    "land_water_fraction": {
        "wflow_v0": None,
        "wflow_v1": "land_water_covered__area_fraction",
    },
    "erosion_soil_detachability": {
        "wflow_v0": "vertical.erosk",
        "wflow_v1": "soil_erosion__rainfall_soil_detachability_factor",
    },
    "erosion_eurosem_exp": {
        "wflow_v0": "vertical.erosspl",
        "wflow_v1": "soil_erosion__eurosem_exponent",
    },
    "erosion_usle_k": {
        "wflow_v0": "vertical.usleK",
        "wflow_v1": "soil_erosion__usle_k_factor",
        "hydromt_name": "usle_k",
    },
    "erosion_usle_c": {
        "wflow_v0": "vertical.usleC",
        "wflow_v1": "soil_erosion__usle_c_factor",
        "hydromt_name": "usle_c",
    },
    "erosion_answers_sheet_factor": {
        "wflow_v0": "vertical.erosov",
        "wflow_v1": "soil_erosion__answers_overland_flow_factor",
    },
    "erosion_answers_rain_factor": {
        "wflow_v0": None,
        "wflow_v1": "soil_erosion__answers_rainfall_factor",
    },
    # soil particles
    "soil_clay_fraction": {
        "wflow_v0": None,
        "wflow_v1": "soil_clay__mass_fraction",
    },
    "soil_silt_fraction": {
        "wflow_v0": None,
        "wflow_v1": "soil_silt__mass_fraction",
    },
    "soil_sand_fraction": {
        "wflow_v0": None,
        "wflow_v1": "soil_sand__mass_fraction",
    },
    "soil_sagg_fraction": {
        "wflow_v0": None,
        "wflow_v1": "soil_small_aggregates__mass_fraction",
    },
    "soil_lagg_fraction": {
        "wflow_v0": None,
        "wflow_v1": "soil_large_aggregates__mass_fraction",
    },
    # land transport
    "land_govers_c": {
        "wflow_v0": None,
        "wflow_v1": "land_surface_water_sediment__govers_transport_capacity_coefficient",  # noqa: E501
    },
    "land_govers_n": {
        "wflow_v0": None,
        "wflow_v1": "land_surface_water_sediment__govers_transport_capacity_exponent",
    },
    "soil_sediment_d50": {
        "wflow_v0": None,
        "wflow_v1": "land_surface_sediment__median_diameter",
    },
    # river transport
    "sediment_density": {
        "wflow_v0": "lateral.river.rhos",
        "wflow_v1": "sediment__particle_density",
    },
    "river_sediment_d50": {
        "wflow_v0": "lateral.river.d50engelund",
        "wflow_v1": None,
    },
    "river_bagnold_c": {
        "wflow_v0": "lateral.river.cbagnold",
        "wflow_v1": "river_water_sediment__bagnold_transport_capacity_coefficient",
    },
    "river_bagnold_exp": {
        "wflow_v0": "lateral.river.ebagnold",
        "wflow_v1": "river_water_sediment__bagnold_transport_capacity_exponent",
    },
    "river_kodatie_a": {
        "wflow_v0": None,
        "wflow_v1": "river_water_sediment__kodatie_transport_capacity_a_coefficient",
    },
    "river_kodatie_b": {
        "wflow_v0": None,
        "wflow_v1": "river_water_sediment__kodatie_transport_capacity_b_coefficient",
    },
    "river_kodatie_c": {
        "wflow_v0": None,
        "wflow_v1": "river_water_sediment__kodatie_transport_capacity_c_coefficient",
    },
    "river_kodatie_d": {
        "wflow_v0": None,
        "wflow_v1": "river_water_sediment__kodatie_transport_capacity_d_coefficient",
    },
    "river_bed_sediment_d50": {
        "wflow_v0": "lateral.river.d50",
        "wflow_v1": "river_bottom_and_bank_sediment__median_diameter",
    },
    "river_bed_clay_fraction": {
        "wflow_v0": "lateral.river.fclayriv",
        "wflow_v1": "river_bottom_and_bank_clay__mass_fraction",
    },
    "river_bed_silt_fraction": {
        "wflow_v0": "lateral.river.fsiltriv",
        "wflow_v1": "river_bottom_and_bank_silt__mass_fraction",
    },
    "river_bed_sand_fraction": {
        "wflow_v0": "lateral.river.fsandriv",
        "wflow_v1": "river_bottom_and_bank_sand__mass_fraction",
    },
    "river_bed_gravel_fraction": {
        "wflow_v0": "lateral.river.fgravelriv",
        "wflow_v1": "river_bottom_and_bank_gravel__mass_fraction",
    },
    "sediment_clay_dm": {
        "wflow_v0": None,
        "wflow_v1": "clay__mean_diameter",
    },
    "sediment_silt_dm": {
        "wflow_v0": None,
        "wflow_v1": "silt__mean_diameter",
    },
    "sediment_sand_dm": {
        "wflow_v0": None,
        "wflow_v1": "sand__mean_diameter",
    },
    "sediment_sagg_dm": {
        "wflow_v0": None,
        "wflow_v1": "sediment_small_aggregates__mean_diameter",
    },
    "sediment_lagg_dm": {
        "wflow_v0": None,
        "wflow_v1": "sediment_large_aggregates__mean_diameter",
    },
    "sediment_gravel_dm": {
        "wflow_v0": None,
        "wflow_v1": "gravel__mean_diameter",
    },
}

WFLOW_SEDIMENT_STATES_NAMES: dict[str, dict[str, str | None]] = {
    "river_clay_load": {
        "wflow_v0": "lateral.river.clayload",
        "wflow_v1": "river_water_clay__mass",
    },
    "river_bed_clay_store": {
        "wflow_v0": "lateral.river.claystore",
        "wflow_v1": "river_bed_clay__mass",
    },
    "river_clay_flux": {
        "wflow_v0": "lateral.river.outclay",
        "wflow_v1": "river_water_clay__mass_flow_rate",
    },
    "river_gravel_load": {
        "wflow_v0": "lateral.river.gravload",
        "wflow_v1": "river_water_gravel__mass",
    },
    "river_bed_gravel_store": {
        "wflow_v0": "lateral.river.gravstore",
        "wflow_v1": "river_bed_gravel__mass",
    },
    "river_gravel_flux": {
        "wflow_v0": "lateral.river.outgrav",
        "wflow_v1": "river_water_gravel__mass_flow_rate",
    },
    "river_lagg_load": {
        "wflow_v0": "lateral.river.laggload",
        "wflow_v1": "river_water_large_aggregates__mass",
    },
    "river_bed_lagg_store": {
        "wflow_v0": "lateral.river.laggstore",
        "wflow_v1": "river_bed_large_aggregates__mass",
    },
    "river_lagg_flux": {
        "wflow_v0": "lateral.river.outlagg",
        "wflow_v1": "river_water_large_aggregates__mass_flow_rate",
    },
    "river_sagg_load": {
        "wflow_v0": "lateral.river.saggload",
        "wflow_v1": "river_water_small_aggregates__mass",
    },
    "river_bed_sagg_store": {
        "wflow_v0": "lateral.river.saggstore",
        "wflow_v1": "river_bed_small_aggregates__mass",
    },
    "river_sagg_flux": {
        "wflow_v0": "lateral.river.outsagg",
        "wflow_v1": "river_water_small_aggregates__mass_flow_rate",
    },
    "river_sand_load": {
        "wflow_v0": "lateral.river.sandload",
        "wflow_v1": "river_water_sand__mass",
    },
    "river_bed_sand_store": {
        "wflow_v0": "lateral.river.sandstore",
        "wflow_v1": "river_bed_sand__mass",
    },
    "river_sand_flux": {
        "wflow_v0": "lateral.river.outsand",
        "wflow_v1": "river_water_sand__mass_flow_rate",
    },
    "river_silt_load": {
        "wflow_v0": "lateral.river.siltload",
        "wflow_v1": "river_water_silt__mass",
    },
    "river_bed_silt_store": {
        "wflow_v0": "lateral.river.siltstore",
        "wflow_v1": "river_bed_silt__mass",
    },
    "river_silt_flux": {
        "wflow_v0": "lateral.river.outsilt",
        "wflow_v1": "river_water_silt__mass_flow_rate",
    },
}


def _create_hydromt_wflow_mapping(
    hydromt_dict: dict,
    model_dict: dict,
    config_dict: dict,
) -> tuple[dict, dict]:
    """
    Create dictionaries to convert from hydromt/Wflow names to staticmaps names.

    The first dictionary will be used to rename from hydromt convention
    name into name in staticmaps/forcing.
    The second dictionary will be used to link the right name
    in staticmaps/forcing files to the right Wflow.jl variables in the toml file.

    Parameters
    ----------
    hydromt_dict : dict
        Dictionary of default hydromt names and name in staticmaps for the variables
        that are not needed by Wflow but needed for model building.
    model_dict : dict
        Dictionary of default name in staticmaps and a potential "hydromt_name" key if
        some of the variables have a convention on name in hydromt. Also contains
        the wflow variable names to be able to update the default names based on the
        config.
    config_dict : dict
        Dictionary of the current model config to update from the default name in
        staticmaps to the actual name in staticmaps.

    Returns
    -------
    mapping_hydromt : dict
        Dictionary of the mapping from hydromt names to staticmaps names.
    mapping_wflow : dict
        Dictionary of the mapping from staticmaps names to Wflow variable names.
    """
    wflow_version = "wflow_v1"  # variable version to use for Wflow.jl

    # First dictionary
    # name_in_staticmaps : name_in_hydromt
    # Instantiate the mapping with default names (ie non wflow variables)
    mapping_inv = {v: k for k, v in hydromt_dict.items()}

    # Second dictionary
    wflow_names = dict()  # wflow_variable: name_in_staticmaps

    # Go through the wflow variables and add them if hydromt name is not None
    for staticmap_name, staticmap_mapping in model_dict.items():
        if isinstance(staticmap_mapping, dict):
            # use hydromt_name if it is available, use default staticmap name otherwise
            mapping_inv[staticmap_name] = staticmap_mapping.get(
                "hydromt_name", staticmap_name
            )
            if wflow_version in staticmap_mapping:
                wflow_names[staticmap_mapping.get(wflow_version)] = staticmap_name

    # Update with the TOML
    # Check if wflow v0 config then do not update
    # (wflow_v0 only support upgrade function so this step is not needed)
    is_wflow_v0 = _check_wflow_version(config_dict)

    if "input" in config_dict and not is_wflow_v0:
        variable_types = ["", "forcing", "cyclic", "static"]
        for var_type in variable_types:
            if var_type != "" and var_type not in config_dict["input"]:
                continue
            # Variables in the general input section
            if var_type == "":
                vars_dict = {}
                for input_var in WFLOW_VARS_IN_INPUT:
                    if input_var in config_dict["input"]:
                        vars_dict[input_var] = config_dict["input"][input_var]
            # Others (static / cyclic / forcing)
            else:
                vars_dict = config_dict["input"][var_type]
            # Go through the variables that can be renamed
            for var_name, new_name in vars_dict.items():
                # Get the old name
                old_name = wflow_names.get(var_name)
                # If they are different update the mapping
                if isinstance(new_name, dict):
                    if "value" in new_name:
                        # Do not update
                        new_name = None
                    elif "netcdf_variable_name" in new_name:
                        new_name = new_name["netcdf_variable_name"]
                if new_name is not None and old_name != new_name:
                    # Update the mapping with the new name
                    mapping_inv[new_name] = mapping_inv.get(old_name, old_name)
                    # Remove the old name from the mapping
                    mapping_inv.pop(old_name, None)
                    # Update wflow_names with the new name
                    wflow_names[var_name] = new_name

    # Invert mapping to get hydromt_name: staticmap_name
    mapping_hydromt = {
        hydromt_name: staticmap_name
        for (staticmap_name, hydromt_name) in mapping_inv.items()
    }

    # # Get a mapping of staticmap_name: wflow_variable
    mapping_wflow = {
        staticmap_name: wflow_var for wflow_var, staticmap_name in wflow_names.items()
    }

    return mapping_hydromt, mapping_wflow


def _create_hydromt_wflow_mapping_sbm(config: dict) -> tuple[dict, dict]:
    """
    Create a dictionary to convert from hydromt names to wflow sbm input names.

    These names will be used in staticmaps/forcing and linked to the right wflow
    internal variables in the toml.
    """
    return _create_hydromt_wflow_mapping(HYDROMT_NAMES_DEFAULT, WFLOW_NAMES, config)


def _create_hydromt_wflow_mapping_sediment(config: dict) -> tuple[dict, dict]:
    """
    Create a dictionary to convert from hydromt names to wflow sediment input names.

    These names will be used in staticmaps/forcing and linked to the right wflow
    internal variables in the toml.
    """
    return _create_hydromt_wflow_mapping(
        HYDROMT_NAMES_DEFAULT_SEDIMENT, WFLOW_SEDIMENT_NAMES, config
    )


def _check_wflow_version(config_dict: dict) -> bool:
    """
    Check if the config is for Wflow v0 or v1.

    Parameters
    ----------
    config_dict : dict
        Dictionary of the current model config.

    Returns
    -------
    bool
        True if the config is for Wflow v0, False otherwise.
    """
    _is_wflow_v0 = False
    if "starttime" in config_dict:  # first check
        _is_wflow_v0 = True
    # second check for safety as starttime is not present for fews_run
    if "input" in config_dict and "vertical" in config_dict["input"]:
        _is_wflow_v0 = True

    return _is_wflow_v0



"""Basemap workflows for Wflow plugin."""

import logging
import math
from typing import Optional

import geopandas as gpd
import numpy as np
import pyflwdir
import xarray as xr
from hydromt import DataCatalog
from hydromt.gis import flw
from hydromt.model.processes.basin_mask import get_basin_geometry
from hydromt.model.processes.region import (
    _parse_region_value,
    parse_region_bbox,
    parse_region_geom,
)
from pyflwdir import core_conversion, core_d8, core_ldd

logger = logging.getLogger(f"hydromt.{__name__}")

__all__ = ["hydrography", "topography", "parse_region"]


def hydrography(
    ds: xr.Dataset,
    res: float,
    xy: Optional[gpd.GeoDataFrame] = None,
    upscale_method: str = "ihu",
    flwdir_name: str = "flwdir",
    uparea_name: str = "uparea",
    basins_name: str = "basins",
    strord_name: str = "strord",
    ftype: str = "infer",
):
    """Return hydrography maps (see list below) and FlwdirRaster object.

    Based on gridded flow direction and elevation data input.

    The output maps are:

    - flwdir : flow direction [-]
    - basins : basin map [-]
    - uparea : upstream area [km2]
    - strord : stream order [-]

    If the resolution is lower than the source resolution, the flow direction data is
    upscaled and river length and slope are based on subgrid flow paths and
    the following maps are added:

    - subare : contributing area to each subgrid outlet pixel \
(unit catchment area) [km2]
    - subelv : elevation at subgrid outlet pixel [m+REF]

    Parameters
    ----------
    ds : xarray.DataArray
        Dataset containing gridded flow direction and elevation data.
    res : float
        output resolution
    xy : geopandas.GeoDataFrame, optional
        Subbasin pits. Only required when upscaling a subbasin.
    upscale_method : {'ihu', 'eam', 'dmm'}
        Upscaling method for flow direction data, by default 'ihu', see [1]_
    ftype : {'d8', 'ldd', 'nextxy', 'nextidx', 'infer'}, optional
        name of flow direction type, infer from data if 'infer', by default is 'infer'
    flwdir_name, elevtn_name, uparea_name, basins_name, strord_name : str, optional
        Name of flow direction [-], elevation [m], upstream area [km2], basin index [-]
        and stream order [-] variables in ds

    Returns
    -------
    ds_out : xarray.DataArray
        Dataset containing gridded hydrography data
    flwdir_out : pyflwdir.FlwdirRaster
        Flow direction raster object.

    References
    ----------
    .. [1] Eilander et al. (2021). A hydrography upscaling method for scale-invariant \
parametrization of distributed hydrological models.
           Hydrology and Earth System Sciences, 25(9), 5287–5313. https://doi.org/10.5194/hess-25-5287-2021

    See Also
    --------
    pyflwdir.FlwdirRaster.upscale_flwdir

    """
    # TODO add check if flwdir in ds, calculate if not
    flwdir = None
    basins = None
    outidx = None
    if "mask" not in ds.coords and xy is None:
        ds.coords["mask"] = xr.Variable(
            dims=ds.raster.dims, data=np.ones(ds.raster.shape, dtype=bool)
        )
    elif "mask" not in ds.coords:
        # NOTE if no subbasin mask is provided calculate it here
        logger.debug(f"Delineate {xy[0].size} subbasin(s).")
        flwdir = flw.flwdir_from_da(ds[flwdir_name], ftype=ftype)
        basins = flwdir.basins(xy=xy).astype(np.int32)
        ds.coords["mask"].data = basins != 0
        if not np.any(ds.coords["mask"]):
            raise ValueError("Delineating subbasins not successful.")
    elif xy is not None:
        # NOTE: this mask is passed on from get_basin_geometry method
        logger.debug("Mask in dataset assumed to represent subbasins.")
    ncells = np.sum(ds["mask"].values)
    scale_ratio = int(np.round(res / ds.raster.res[0]))

    if ncells < 4:
        raise ValueError(
            "(Sub)basin at original resolution should at least consist of two cells on "
            f"each axis and the total number of cells is {ncells}. "
            "Consider using a larger domain or higher spatial resolution. "
            "For subbasin models, consider a (higher) threshold on for example "
            "upstream area or stream order to snap the outlet."
        )
    elif ncells < 100 and scale_ratio > 1:
        logger.warning(
            f"(Sub)basin at original resolution is small and has {ncells} cells. "
            "This may results in errors later when upscaling flow directions. "
            "If so, consider using a larger domain or higher spatial resolution. "
            "For subbasin models, consider a (higher) threshold on for example "
            "upstream area or stream order to snap the outlet."
        )
    else:
        logger.info(f"(Sub)basin at original resolution has {ncells} cells.")

    if scale_ratio > 1:  # upscale flwdir
        if flwdir is None:
            # NOTE initialize with mask is FALSE
            flwdir = flw.flwdir_from_da(ds[flwdir_name], ftype=ftype, mask=False)
        if xy is not None:
            logger.debug("Burn subbasin outlet in upstream area data.")
            if isinstance(xy, gpd.GeoDataFrame):
                assert xy.crs == ds.raster.crs
                xy = xy.geometry.x, xy.geometry.y
            idxs_pit = flwdir.index(*xy)
            flwdir.add_pits(idxs=idxs_pit)
            uparea = ds[uparea_name].values
            uparea.flat[idxs_pit] = uparea.max() + 1.0
            ds[uparea_name].data = uparea
        logger.info(
            f"Upscale flow direction data: {scale_ratio:d}x, {upscale_method} method."
        )
        da_flw, flwdir_out = flw.upscale_flwdir(
            ds,
            flwdir=flwdir,
            scale_ratio=scale_ratio,
            method=upscale_method,
            uparea_name=uparea_name,
            flwdir_name=flwdir_name,
        )
        da_flw.raster.set_crs(ds.raster.crs)
        # make sure x_out and y_out get saved
        ds_out = da_flw.to_dataset().reset_coords(["x_out", "y_out"])
        dims = ds_out.raster.dims
        # find pits within basin mask
        idxs_pit0 = flwdir_out.idxs_pit
        outlon = ds_out["x_out"].values.ravel()
        outlat = ds_out["y_out"].values.ravel()
        sel = {
            ds.raster.x_dim: xr.Variable("yx", outlon[idxs_pit0]),
            ds.raster.y_dim: xr.Variable("yx", outlat[idxs_pit0]),
        }
        outbas_pit = ds.coords["mask"].sel(sel, method="nearest").values
        # derive basins
        if np.any(outbas_pit != 0):
            idxs_pit = idxs_pit0[outbas_pit != 0]
            basins = flwdir_out.basins(idxs=idxs_pit).astype(np.int32)
            ds_out.coords["mask"] = xr.Variable(
                dims=ds_out.raster.dims, data=basins != 0, attrs=dict(_FillValue=0)
            )
        else:
            # This is a patch for basins which are clipped based on bbox or wrong geom
            mask_int = ds["mask"].astype(np.int8)
            mask_int.raster.set_nodata(-1)  # change nodata value
            ds_out.coords["mask"] = mask_int.raster.reproject_like(
                da_flw, method="nearest"
            ).astype(bool)
            basins = ds_out["mask"].values.astype(np.int32)
            logger.warning(
                "The basin delineation might be wrong as no original resolution outlets"
                " are found in the upscaled map."
            )
        ds_out[basins_name] = xr.Variable(dims, basins, attrs=dict(_FillValue=0))
        # calculate upstream area using subgrid ucat cell areas
        outidx = np.where(
            ds_out["mask"], da_flw.coords["idx_out"].values, flwdir_out._mv
        )
        subare = flwdir.ucat_area(outidx, unit="km2")[1]
        uparea = flwdir_out.accuflux(subare)
        attrs = dict(_FillValue=-9999, unit="km2")
        ds_out[uparea_name] = xr.Variable(dims, uparea, attrs=attrs).astype(np.float32)
        # NOTE: subgrid cella area is currently not used in wflow
        ds_out["subare"] = xr.Variable(dims, subare, attrs=attrs).astype(np.float32)
        if "elevtn" in ds:
            subelv = ds["elevtn"].values.flat[outidx]
            subelv = np.where(outidx >= 0, subelv, -9999)
            attrs = dict(_FillValue=-9999, unit="m+REF")
            ds_out["subelv"] = xr.Variable(dims, subelv, attrs=attrs)
        # initiate masked flow dir
        flwdir_out = flw.flwdir_from_da(
            ds_out[flwdir_name], ftype=flwdir.ftype, mask=True
        )
    else:
        # NO upscaling : source resolution equals target resolution
        # NOTE (re-)initialize with mask is TRUE
        ftype = flwdir.ftype if flwdir is not None and ftype == "infer" else ftype
        flwdir = flw.flwdir_from_da(ds[flwdir_name], ftype=ftype, mask=True)
        flwdir_out = flwdir
        ds_out = xr.DataArray(
            name=flwdir_name,
            data=flwdir_out.to_array(),
            coords=ds.raster.coords,
            dims=ds.raster.dims,
            attrs=dict(
                long_name=f"{ftype} flow direction",
                _FillValue=flwdir_out._core._mv,
            ),
        ).to_dataset()
        dims = ds_out.raster.dims
        ds_out.coords["mask"] = xr.Variable(
            dims=dims, data=flwdir_out.mask.reshape(flwdir_out.shape)
        )
        # copy data variables from source if available
        for dvar in [basins_name, uparea_name, strord_name]:
            if dvar in ds.data_vars:
                ds_out[dvar] = xr.where(
                    ds_out["mask"],
                    ds[dvar],
                    ds[dvar].dtype.type(ds[dvar].raster.nodata),
                )
                ds_out[dvar].attrs.update(ds[dvar].attrs)
        # basins
        if basins_name not in ds_out.data_vars:
            if basins is None:
                basins = flwdir_out.basins(idxs=flwdir_out.idxs_pit).astype(np.int32)
            ds_out[basins_name] = xr.Variable(dims, basins, attrs=dict(_FillValue=0))
        else:
            # make sure dtype in ds_out is np.int32
            ds_out[basins_name] = ds_out[basins_name].astype(np.int32)
        # upstream area
        if uparea_name not in ds_out.data_vars:
            uparea = flwdir_out.upstream_area("km2")  # km2
            attrs = dict(_FillValue=-9999, unit="km2")
            ds_out[uparea_name] = xr.Variable(dims, uparea, attrs=attrs).astype(
                np.float32
            )
        # cell area
        # NOTE: subgrid cella area is currently not used in wflow
        subare = ds.raster.area_grid() / 1e6  # km2
        attrs = dict(_FillValue=-9999, unit="km2")
        ds_out["subare"] = xr.Variable(dims, subare, attrs=attrs).astype(np.float32)
    # logging
    npits = flwdir_out.idxs_pit.size
    xy_pit = flwdir_out.xy(flwdir_out.idxs_pit[:5])
    xy_pit_str = ", ".join([f"({x:.5f},{y:.5f})" for x, y in zip(*xy_pit)])
    # stream order
    if strord_name not in ds_out.data_vars:
        logger.info("Derive stream order.")
        strord = flwdir_out.stream_order()
        ds_out[strord_name] = xr.Variable(dims, strord)
        ds_out[strord_name].raster.set_nodata(255)

    # clip to basin extent
    ds_out = ds_out.raster.clip_mask(da_mask=ds_out[basins_name])
    # also mask idx_out coords if present
    if "idx_out" in ds_out:
        ds_out["idx_out"] = ds_out["idx_out"].where(
            ds_out[basins_name], ds_out["idx_out"].raster.nodata
        )

    ds_out.raster.set_crs(ds.raster.crs)
    logger.info(f"Map shape: {ds_out.raster.shape}; active cells: {flwdir_out.ncells}.")
    logger.info(f"Outlet coordinates ({len(xy_pit[0])}/{npits}): {xy_pit_str}.")
    if np.any(np.asarray(ds_out.raster.shape) == 1):
        raise ValueError(
            "The output extent at model resolution should at least consist of two "
            "cells on each axis. Consider using a larger domain or higher spatial "
            "resolution. For subbasin models, consider a (higher) threshold to snap "
            "the outlet."
        )

    # Convert flow direction from d8 to ldd format
    flwdir_data = ds_out["flwdir"].values.astype(np.uint8)  # force dtype
    # if d8 convert to ldd
    if core_d8.isvalid(flwdir_data):
        data = core_conversion.d8_to_ldd(flwdir_data)
        da_flwdir = xr.DataArray(
            name="flwdir",
            data=data,
            coords=ds_out.raster.coords,
            dims=ds_out.raster.dims,
            attrs=dict(
                long_name="ldd flow direction",
                _FillValue=core_ldd._mv,
            ),
        )
        ds_out["flwdir"] = da_flwdir
    return ds_out, flwdir_out


def topography(
    ds: xr.Dataset,
    ds_like: xr.Dataset,
    elevtn_name: str = "elevtn",
    lndslp_name: str = "lndslp",
    method: str = "average",
):
    """Return topography maps (see list below) at model resolution.

    Based on gridded elevation data input.

    The following topography maps are calculated:

    - elevtn : average elevation [m]
    - lndslp : average land surface slope [m/m]

    Parameters
    ----------
    ds : xarray.DataArray
        Dataset containing gridded flow direction and elevation data.
    ds_like : xarray.DataArray
        Dataset at model resolution.
    elevtn_name, lndslp_name : str, optional
        Name of elevation [m] and land surface slope [m/m] variables in ds
    method: str, optional
        Resample method used to reproject the input data, by default "average"

    Returns
    -------
    ds_out : xarray.DataArray
        Dataset containing gridded hydrography data
    flwdir_out : pyflwdir.FlwdirRaster
        Flow direction raster object.

    See Also
    --------
    pyflwdir.dem.slope
    """
    logger.info("Derive elevation and slope.")
    if lndslp_name not in ds.data_vars:
        logger.info(f"Slope map {lndslp_name} not found: derive from elevation map.")
        crs = ds[elevtn_name].raster.crs
        nodata = ds[elevtn_name].raster.nodata
        ds[lndslp_name] = xr.Variable(
            dims=ds.raster.dims,
            data=pyflwdir.dem.slope(
                elevtn=ds[elevtn_name].values,
                nodata=nodata,
                latlon=crs is not None and crs.to_epsg() == 4326,
                transform=ds[elevtn_name].raster.transform,
            ),
        )
        ds[lndslp_name].raster.set_nodata(nodata)
    # clip or reproject if non-identical grids
    ds_out = ds[[elevtn_name, lndslp_name]].raster.reproject_like(ds_like, method)
    # Mask possible negative values for slope
    ds_out["lndslp"] = np.maximum(ds_out["lndslp"], 0.0)
    ds_out[elevtn_name].attrs.update(unit="m")
    ds_out[lndslp_name].attrs.update(unit="m.m-1")
    return ds_out


def parse_region(
    data_catalog: DataCatalog,
    region: dict,
    hydrography_fn: str | xr.Dataset,
    resolution: float | int = 1 / 120.0,
    basin_index_fn: str | xr.Dataset | None = None,
) -> tuple[dict[str, gpd.GeoDataFrame], Optional[np.ndarray], xr.Dataset]:
    """Parse the region dictionary to get basin geometry and coordinates."""
    ds_org = data_catalog.get_rasterdataset(hydrography_fn)
    if ds_org is None:
        raise ValueError(f"hydrography_fn {hydrography_fn} not found in data catalog.")

    # Check on resolution
    hydrography_resolution = ds_org.raster.res[0]
    scale_ratio = resolution / hydrography_resolution

    if math.isclose(scale_ratio, 1):
        pass
    elif scale_ratio < 0.75:
        raise ValueError(
            f"Model resolution {resolution} should be larger than the "
            f"{hydrography_fn} resolution {hydrography_resolution}."
        )
    elif 0.75 < scale_ratio < 1.25:
        logger.warning(
            f"Model resolution {resolution} does not match the hydrography "
            f"resolution {hydrography_resolution}. This might lead to unexpected "
            f"results, using hydrography resolution instead: "
            f"{hydrography_resolution}."
        )
        resolution = hydrography_resolution
    else:
        if ds_org.raster.crs.is_geographic:
            if resolution > 1:  # 111 km
                raise ValueError(
                    f"The model resolution {resolution} should be smaller than 1 "
                    "degree (111km) for geographic coordinate systems. "
                    "Make sure you provided res in degree rather than in meters."
                )

    # get basin geometry and clip data
    kind = next(iter(region))
    xy = None
    if kind in ["basin", "subbasin"]:
        # parse_region_basin does not return xy, only geom...
        # should be fixed in core
        region_kwargs = _parse_region_value(
            region.pop(kind),
            data_catalog=data_catalog,
        )
        region_kwargs.update(region)
        if basin_index_fn is not None:
            bas_index = data_catalog.get_source(basin_index_fn)
        else:
            bas_index = None
        geom, xy = get_basin_geometry(
            ds=ds_org,
            kind=kind,
            basin_index=bas_index,
            **region_kwargs,
        )
    elif kind == "bbox":
        logger.warning(
            "Kind 'bbox' for the region is not recommended as it can lead "
            "to mistakes in the catchment delineation. Use carefully."
        )
        geom = parse_region_bbox(region)
    elif kind == "geom":
        logger.warning(
            "Kind 'geom' for the region is not recommended as it can lead "
            "to mistakes in the catchment delineation. Use carefully."
        )
        geom = parse_region_geom(region)
    else:
        raise ValueError(
            f"wflow region kind not understood or supported: {kind}. "
            "Use 'basin', 'subbasin', 'bbox' or 'geom'."
        )

    if geom is not None and geom.crs is None:
        raise ValueError("wflow region geometry has no CRS")

    # Set the basins geometry
    ds_org = ds_org.raster.clip_geom(geom, align=resolution, buffer=10)
    ds_org.coords["mask"] = ds_org.raster.geometry_mask(geom)

    geometries = {}
    if not math.isclose(scale_ratio, 1):
        geometries["basins_highres"] = geom

    return geometries, xy, ds_org



"""Implement the Wflow Sediment model class."""

import logging
import tomllib
from pathlib import Path

import geopandas as gpd
import numpy as np
import pandas as pd
import xarray as xr
from hydromt import hydromt_step
from hydromt.error import NoDataStrategy

import hydromt_wflow.utils as utils
from hydromt_wflow import workflows
from hydromt_wflow.naming import _create_hydromt_wflow_mapping_sediment
from hydromt_wflow.version_upgrade import (
    convert_reservoirs_to_wflow_v1_sediment,
    convert_to_wflow_v1_sediment,
)
from hydromt_wflow.wflow_base import WflowBaseModel

__all__ = ["WflowSedimentModel"]
__hydromt_eps__ = ["WflowSedimentModel"]
logger = logging.getLogger(f"hydromt.{__name__}")


class WflowSedimentModel(WflowBaseModel):
    """The wflow sediment model class, a subclass of WflowBaseModel."""

    name: str = "wflow_sediment"

    def __init__(
        self,
        root: str | None = None,
        config_filename: str | None = None,
        mode: str = "w",
        data_libs: list | str = [],
        **catalog_keys,
    ):
        super().__init__(
            root=root,
            config_filename=config_filename,
            mode=mode,
            data_libs=data_libs,
            **catalog_keys,
        )
        # Update compared to wflow sbm
        self._MAPS, self._WFLOW_NAMES = _create_hydromt_wflow_mapping_sediment(
            self.config.data
        )

    @hydromt_step
    def setup_rivers(
        self,
        hydrography_fn: str | xr.Dataset,
        river_geom_fn: str | gpd.GeoDataFrame | None = None,
        river_upa: float = 30,
        slope_len: float = 2e3,
        min_rivlen_ratio: float = 0.0,
        min_rivwth: float = 30,
        smooth_len: float = 5e3,
        output_names: dict = {
            "river_location__mask": "river_mask",
            "river__length": "river_length",
            "river__width": "river_width",
            "river__slope": "river_slope",
        },
    ):
        """
        Set all river parameter maps.

        The river mask is defined by all cells with a minimum upstream area threshold
        ``river_upa`` [km2].

        The river length is defined as the distance from the subgrid outlet pixel to
        the next upstream subgrid outlet pixel. The ``min_rivlen_ratio`` is the minimum
        global river length to avg. cell resolution ratio and is used as a threshold in
        window based smoothing of river length.

        The river slope is derived from the subgrid elevation difference between pixels
        at a half distance ``slope_len`` [m] up-
        and downstream from the subgrid outlet pixel.

        The river manning roughness coefficient is derived based on reclassification
        of the streamorder map using a lookup table ``rivman_mapping_fn``.

        The river width is derived from the nearest river segment in ``river_geom_fn``.
        Data gaps are filled by the nearest valid upstream value and averaged along
        the flow directions over a length ``smooth_len`` [m].

        Adds model layers:

            * **river_mask** map: river mask [-]
            * **river_length** map: river length [m]
            * **river_width** map: river width [m]
            * **river_slope** map: river slope [m/m]
            * **rivers** geom: river vector based on river_mask

        Required setup methods:

        * :py:meth:`~WflowSedimentModel.setup_basemaps`

        Parameters
        ----------
        hydrography_fn : str, Path, xarray.Dataset
            Name of RasterDataset source for hydrography data.
            Must be same as setup_basemaps for consistent results.

            * **Required variables**: 'flwdir' [LLD or D8 or NEXTXY], 'uparea' [km2],
              'elevtn' [m+REF]

            * **Optional variables**: 'rivwth' [m]

        river_geom_fn : str, Path, geopandas.GeoDataFrame, optional
            Name of GeoDataFrame source for river data.

            * **Required variables**: 'rivwth' [m]

        river_upa : float, optional
            Minimum upstream area threshold for the river map [km2]. By default 30.0
        slope_len : float, optional
            Length over which the river slope is calculated [km]. By default 2.0
        min_rivlen_ratio: float, optional
            Ratio of cell resolution used minimum length threshold in a moving
            window based smoothing of river length, by default 0.0
            The river length smoothing is skipped if `min_riverlen_ratio` = 0.
            For details about the river length smoothing,
            see :py:meth:`pyflwdir.FlwdirRaster.smooth_rivlen`
        smooth_len : float, optional
            Length [m] over which to smooth the output river width,
            by default 5e3
        min_rivwth : float, optional
            Minimum river width [m], by default 30.0
        output_names : dict, optional
            Dictionary with output names that will be used in the model netcdf input
            files. Users should provide the Wflow.jl variable name followed by the name
            in the netcdf file.
        """
        super().setup_rivers(
            hydrography_fn=hydrography_fn,
            river_geom_fn=river_geom_fn,
            river_upa=river_upa,
            slope_len=slope_len,
            min_rivlen_ratio=min_rivlen_ratio,
            smooth_len=smooth_len,
            min_rivwth=min_rivwth,
            rivdph_method=None,
            min_rivdph=None,
            output_names=output_names,
        )

    @hydromt_step
    def setup_riverbedsed(
        self,
        bedsed_mapping_fn: str | Path | pd.DataFrame | None = None,
        strord_name: str = "meta_streamorder",
        output_names: dict = {
            "river_bottom_and_bank_sediment__median_diameter": "river_bed_sediment_d50",
            "river_bottom_and_bank_clay__mass_fraction": "river_bed_clay_fraction",
            "river_bottom_and_bank_silt__mass_fraction": "river_bed_silt_fraction",
            "river_bottom_and_bank_sand__mass_fraction": "river_bed_sand_fraction",
            "river_bottom_and_bank_gravel__mass_fraction": "river_bed_gravel_fraction",
            "river_water_sediment__kodatie_transport_capacity_a_coefficient": "river_kodatie_a",  # noqa: E501
            "river_water_sediment__kodatie_transport_capacity_b_coefficient": "river_kodatie_b",  # noqa: E501
            "river_water_sediment__kodatie_transport_capacity_c_coefficient": "river_kodatie_c",  # noqa: E501
            "river_water_sediment__kodatie_transport_capacity_d_coefficient": "river_kodatie_d",  # noqa: E501
        },
    ):
        """Generate sediments based river bed characteristics maps.

        Kodatie transport capacity coefficients can also be derived from such mapping
        table based on the mean sediment diameter of the river bed.

        Adds model layers:

        * **river_bed_sediment_d50** map: median sediment diameter of the river bed [mm]
        * **river_bed_clay_fraction** map: fraction of clay material in the river bed [-]
        * **river_bed_silt_fraction** map: fraction of silt material in the river bed [-]
        * **river_bed_sand_fraction** map: fraction of sand material in the river bed [-]
        * **river_bed_gravel_fraction** map: fraction of gravel material in the river bed [-]
        * **river_kodatie_a** map: Kodatie transport capacity coefficient a [-]
        * **river_kodatie_b** map: Kodatie transport capacity coefficient b [-]
        * **river_kodatie_c** map: Kodatie transport capacity coefficient c [-]
        * **river_kodatie_d** map: Kodatie transport capacity coefficient d [-]

        Parameters
        ----------
        bedsed_mapping_fn : str
            Path to a mapping csv file from streamorder to river bed particles
            characteristics. If None reverts to default values.

            * Required variable: ['strord','river_bed_sediment_d50', \
'river_bed_clay_fraction', 'river_bed_silt_fraction', 'river_bed_sand_fraction', \
'river_bed_gravel_fraction']
            * Optional variable: ['river_kodatie_a', 'river_kodatie_b', \
'river_kodatie_c', 'river_kodatie_d']

        strord_name : str, optional
            Name of the stream order map in the grid, by default 'meta_streamorder'.
        output_names : dict, optional
            Dictionary with output names that will be used in the model netcdf input
            files. Users should provide the Wflow.jl variable name followed by the name
            in the netcdf file.
        """  # noqa: E501
        logger.info("Preparing riverbedsed parameter maps.")
        # check for streamorder
        if self._MAPS["strord"] not in self.staticmaps.data:
            if strord_name not in self.staticmaps.data:
                raise ValueError(
                    f"Streamorder map {strord_name} not found in grid. "
                    "Please run setup_basemaps or update the strord_name argument."
                )
            else:
                self._MAPS["strord"] = strord_name
        # update self._MAPS and self._WFLOW_NAMES with user defined output names
        self._update_naming(output_names)

        # Make river_bed_sediment_d50 map from csv file with mapping between streamorder
        #  and river_bed_sediment_d50 value
        if bedsed_mapping_fn is None:
            fn_map = "riverbedsed_mapping_default"
        else:
            fn_map = bedsed_mapping_fn

        df = self.data_catalog.get_dataframe(fn_map)

        strord = self.staticmaps.data[self._MAPS["strord"]].copy()
        # max streamorder value above which values get the same D50 value
        max_str = df.index[-2]
        nodata = df.index[-1]
        # if streamroder value larger than max_str, assign last value
        strord = strord.where(strord <= max_str, max_str)
        # handle missing value (last row of csv is mapping of nan values)
        strord = strord.where(strord != strord.raster.nodata, nodata)
        strord.raster.set_nodata(nodata)

        ds_riversed = workflows.landuse(
            da=strord,
            ds_like=self.staticmaps.data,
            df=df,
        )

        rmdict = {k: self._MAPS.get(k, k) for k in ds_riversed.data_vars}
        self.staticmaps.set(ds_riversed.rename(rmdict))
        # update config
        self._update_config_variable_name(ds_riversed.rename(rmdict).data_vars)

    @hydromt_step
    def setup_natural_reservoirs(
        self,
        reservoirs_fn: str | Path | gpd.GeoDataFrame,
        overwrite_existing: bool = False,
        duplicate_id: str = "error",
        min_area: float = 1.0,
        output_names: dict = {
            "reservoir_area__count": "reservoir_area_id",
            "reservoir_location__count": "reservoir_outlet_id",
            "reservoir_surface__area": "reservoir_area",
            "reservoir_water_sediment__bedload_trapping_efficiency": "reservoir_trapping_efficiency",  # noqa : E501
        },
        geom_name: str = "meta_natural_reservoirs",
        **kwargs,
    ):
        """Generate maps of natural reservoir areas (lakes) and outlets.

        This method is a specialized version of the `setup_reservoirs` method
        that is specifically designed for lakes, which are natural water bodies
        without artificial dams.

        This means the trapping efficiency is set to 0.0 by default and the output
        staticgeoms will be called meta_natural_reservoirs.geojson by default.

        For a description of the parameters and functionality, see
        py:meth:`setup_reservoirs`.

        Required setup methods:

        * :py:meth:`~WflowSedimentModel.setup_rivers`

        """
        self.setup_reservoirs(
            reservoirs_fn=reservoirs_fn,
            overwrite_existing=overwrite_existing,
            duplicate_id=duplicate_id,
            min_area=min_area,
            trapping_default=0.0,  # lakes have no trapping efficiency
            output_names=output_names,
            geom_name=geom_name,
            source_kwargs=kwargs,
        )

    @hydromt_step
    def setup_reservoirs(
        self,
        reservoirs_fn: str | Path | gpd.GeoDataFrame,
        overwrite_existing: bool = False,
        duplicate_id: str = "error",
        min_area: float = 1.0,
        trapping_default: float = 1.0,
        output_names: dict = {
            "reservoir_area__count": "reservoir_area_id",
            "reservoir_location__count": "reservoir_outlet_id",
            "reservoir_surface__area": "reservoir_area",
            "reservoir_water_sediment__bedload_trapping_efficiency": "reservoir_trapping_efficiency",  # noqa : E501
        },
        geom_name: str = "meta_reservoirs",
        **kwargs,
    ):
        """Generate maps of reservoir areas and outlets.

        Also generates parameters with average reservoir area,
        and trapping efficiency for large particles.

        The data is generated from features with ``min_area`` [km2] (default is 1 km2)
        from a database with reservoir geometry, IDs and metadata.

        Adds model layers:

        * **reservoir_area_id** map: reservoir IDs [-]
        * **reservoir_outlet_id** map: reservoir IDs at outlet locations [-]
        * **reservoir_area** map: reservoir area [m2]
        * **reservoir_trapping_efficiency** map: reservoir bedload trapping efficiency
          coefficient [-] (0 for natural lakes, 0-1 depending on the type of dam)
        * **meta_reservoirs** geom: polygon with reservoirs and parameters
        * **reservoirs** geom: polygon with all reservoirs as in the model

        Required setup methods:

        * :py:meth:`~WflowSedimentModel.setup_rivers`

        Parameters
        ----------
        reservoirs_fn : str
            Name of data source for reservoir parameters, see ``data/data_sources.yml``.

            * Required variables: ``['waterbody_id', 'Area_avg']``

            * Optional variables: ``['reservoir_trapping_efficiency']``

        overwrite_existing : bool, optional
            If True, overwrite existing reservoirs in the model grid. Default is False.
        duplicate_id : {"error", "skip"}, optional
            Action to take if duplicate reservoir IDs are found when merging with
            existing reservoirs. Options are ``"error"`` (default) to raise an error,
            or ``"skip"`` to skip adding new reservoirs.
        min_area : float, optional
            Minimum reservoir area threshold [km2]. Default is 1.0.
        trapping_default : float, optional
            Default trapping efficiency coefficient for large particles [0-1].
            Default is 1.0, meaning 100% of large particles (sand to gravel)
            are trapped (e.g. in a gravity dam). For other dam types, the
            natural deposition rates from Camp are used.
        output_names : dict, optional
            Dictionary with output names to be used in the model NetCDF input
            files. Keys should be Wflow.jl variable names, values the names in
            the NetCDF file.
        geom_name : str, optional
            Name of the reservoirs geometry in the ``staticgeoms`` folder.
            Default is ``"meta_reservoirs"`` (for meta_reservoirs.geojson).
        kwargs : dict, optional
            Additional keyword arguments passed to
            ``hydromt.DataCatalog.get_rasterdataset()``.
        """
        # retrieve data for basin
        logger.info("Preparing reservoir maps.")
        kwargs.setdefault("predicate", "contains")
        gdf_res = self.data_catalog.get_geodataframe(
            reservoirs_fn,
            geom=self.basins_highres,
            handle_nodata=NoDataStrategy.IGNORE,
            source_kwargs=kwargs,
        )
        # Skip method if no data is returned
        if gdf_res is None:
            logger.info("Skipping method, as no data has been found")
            return

        # Derive reservoir area and outlet maps
        ds_res, gdf_res = workflows.reservoir_id_maps(
            gdf=gdf_res,
            ds_like=self.staticmaps.data,
            min_area=min_area,
            uparea_name=self._MAPS["uparea"],
        )
        if ds_res is None:
            # No reservoir of sufficient size found
            return
        self._update_naming(output_names)

        # Add default trapping efficiency coefficient if not in data source
        if "reservoir_trapping_efficiency" not in gdf_res.columns:
            gdf_res["reservoir_trapping_efficiency"] = trapping_default
        # add reservoirs parameters to grid
        gdf_points = gpd.GeoDataFrame(
            gdf_res[["waterbody_id", "Area_avg", "reservoir_trapping_efficiency"]],
            geometry=gpd.points_from_xy(gdf_res.xout, gdf_res.yout),
        )
        ds_res["reservoir_area"] = self.staticmaps.data.raster.rasterize(
            gdf_points, col_name="Area_avg", dtype="float32", nodata=-999
        )
        ds_res["reservoir_trapping_efficiency"] = self.staticmaps.data.raster.rasterize(
            gdf_points,
            col_name="reservoir_trapping_efficiency",
            dtype="float32",
            nodata=-999,
        )

        # merge with existing reservoirs
        if (
            not overwrite_existing
            and self._MAPS["reservoir_area"] in self.staticmaps.data
        ):
            inv_rename = {
                v: k
                for k, v in self._MAPS.items()
                if v in self.staticmaps.data.data_vars
            }
            ds_res = workflows.reservoirs.merge_reservoirs_sediment(
                ds_res,
                self.staticmaps.data.rename(inv_rename),
                duplicate_id=duplicate_id,
            )
            # Check if ds_res is None ie duplicate IDs
            if ds_res is None:
                logger.warning(
                    "Duplicate reservoir IDs found. Skip adding the new reservoirs."
                )
                return

        # add to grid
        rmdict = {k: self._MAPS.get(k, k) for k in ds_res.data_vars}
        self.staticmaps.set(ds_res.rename(rmdict))

        # write reservoirs with attr tables to static geoms.
        self.geoms.set(gdf_res.rename({"Area_avg": "reservoir_area"}), name=geom_name)
        # Prepare a combined geoms of all reservoirs
        gdf_res_all = workflows.reservoirs.create_reservoirs_geoms_sediment(
            ds_res.rename(rmdict),
        )
        self.geoms.set(gdf_res_all, name="reservoirs")

        # Reservoir settings in the toml to update
        self.config.set("model.reservoir__flag", True)
        for dvar in ds_res.data_vars:
            if dvar in ["reservoir_area_id", "reservoir_outlet_id"]:
                self._update_config_variable_name(self._MAPS[dvar], data_type=None)
            elif dvar in self._WFLOW_NAMES:
                self._update_config_variable_name(self._MAPS[dvar])

    @hydromt_step
    def setup_lulcmaps(
        self,
        lulc_fn: str | Path | xr.DataArray,
        *,
        lulc_mapping_fn: str | Path | pd.DataFrame | None = None,
        planted_forest_fn: str | Path | gpd.GeoDataFrame | None = None,
        lulc_vars: list[str] = [
            "landuse",
            "soil_compacted_fraction",
            "erosion_usle_c",
            "land_water_fraction",
        ],
        planted_forest_c: float = 0.0881,
        orchard_name: str = "Orchard",
        orchard_c: float = 0.2188,
        output_names_suffix: str | None = None,
    ):
        """Derive several wflow maps based on landuse-landcover (LULC) data.

        Lookup table `lulc_mapping_fn` columns are converted to lulc classes model
        parameters based on literature. The data is remapped at its original resolution
        and then resampled to the model resolution using the average value, unless noted
        differently.

        Currently, if `lulc_fn` is set to the "vito", "globcover", "esa_worldcover"
        "corine" or "glmnco", default lookup tables are available and will be used if
        `lulc_mapping_fn` is not provided.

        The USLE C factor map can be refined for planted forests using the planted
        forest data source. The planted forest data source is a polygon layer with
        planted forest polygons and optionally a column with the forest type to
        identify orchards. The default value for orchards is 0.2188, the default value
        for other planted forests is 0.0881.

        Adds model layers:

            * **landuse** map: Landuse class [-]
                Original source dependent LULC class, resampled using nearest neighbour.
            * **erosion_usle_c** map: Cover management factor from the USLE equation [-]
            * **soil_compacted_fraction** map: The fraction of compacted or urban area
                per grid cell [-]
            * **land_water_fraction** map: The fraction of water covered area per
                grid cell [-]

        Required setup methods:

        * :py:meth:`~WflowSedimentModel.setup_basemaps`

        Parameters
        ----------
        lulc_fn : str, Path, xarray.DataArray
            Name of RasterDataset source.
        lulc_mapping_fn : str, Path, pd.DataFrame, None, optional
            Path to a mapping csv file from landuse in source name to parameter values
            in lulc_vars.
        planted_forest_fn : str, Path, gpd.GeoDataFrame, None, optional
            GeoDataFrame source with polygons of planted forests.

            * Optional variable: ["forest_type"]

        lulc_vars : list[str], optional
            List of landuse parameters to prepare.
            The names are the columns of the mapping file.
            Can be a subset of: ["landuse", "soil_compacted_fraction", "erosion_usle_c",
            "land_water_fraction"]
        planted_forest_c : float, optional
            Value of USLE C factor for planted forest, by default 0.0881.
        orchard_name : str, optional
            Name of orchard landuse class in the "forest_type" column of
            ``planted_forest_fn``, by default "Orchard".
        orchard_c : float, optional
            Value of USLE C factor for orchards, by default 0.2188.
        output_names_suffix : str, optional
            Suffix to be added to the output names to avoid having to rename all the
            columns of the mapping tables. For example if the suffix is "vito", all
            variables in lulc_vars will be renamed to "landuse_vito",
            "erosion_usle_c_vito", etc.

        See Also
        --------
        workflows.setup_lulcmaps_from_vector
        workflows.add_planted_forest_to_landuse
        """
        # Prepare all default parameters
        super().setup_lulcmaps(
            lulc_fn=lulc_fn,
            lulc_mapping_fn=lulc_mapping_fn,
            lulc_vars=lulc_vars,
            output_names_suffix=output_names_suffix,
        )

        # If available, improve USLE C map with planted forest data
        if self._MAPS["usle_c"] in lulc_vars and planted_forest_fn is not None:
            # Read forest data
            planted_forest = self.data_catalog.get_geodataframe(
                planted_forest_fn,
                geom=self.basins,
                buffer=1,
                predicate="intersects",
                handle_nodata="IGNORE",
            )
            if planted_forest is None:
                logger.warning("No Planted forest data found within domain.")
                return
            rename_dict = {
                v: k
                for k, v in self._MAPS.items()
                if v in self.staticmaps.data.data_vars
            }
            usle_c = workflows.add_planted_forest_to_landuse(
                planted_forest,
                self.staticmaps.data.rename(rename_dict),
                planted_forest_c=planted_forest_c,
                orchard_name=orchard_name,
                orchard_c=orchard_c,
            )

            # Add to grid
            self.staticmaps.set(usle_c, name=self._MAPS["usle_c"])

    @hydromt_step
    def setup_lulcmaps_from_vector(
        self,
        lulc_fn: str | gpd.GeoDataFrame,
        *,
        lulc_mapping_fn: str | Path | pd.DataFrame | None = None,
        planted_forest_fn: str | Path | gpd.GeoDataFrame | None = None,
        lulc_vars: list[str] = [
            "landuse",
            "soil_compacted_fraction",
            "erosion_usle_c",
            "land_water_fraction",
        ],
        lulc_res: float | int | None = None,
        all_touched: bool = False,
        buffer: int = 1000,
        save_raster_lulc: bool = False,
        planted_forest_c: float = 0.0881,
        orchard_name: str = "Orchard",
        orchard_c: float = 0.2188,
        output_names_suffix: str | None = None,
    ):
        """
        Derive several wflow maps based on vector landuse-landcover (LULC) data.

        The vector lulc data is first rasterized to a raster map at the model resolution
        or at a higher resolution specified in ``lulc_res`` (recommended).

        Lookup table `lulc_mapping_fn` columns are converted to lulc classes model
        parameters based on literature. The data is remapped at its original resolution
        and then resampled to the model resolution using the average value, unless noted
        differently.

        The USLE C factor map can be refined for planted forests using the planted
        forest data source. The planted forest data source is a polygon layer with
        planted forest polygons and optionally a column with the forest type to
        identify orchards. The default value for orchards is 0.2188, the default value
        for other planted forests is 0.0881.

        Adds model layers:

            * **landuse** map: Landuse class [-]
                Original source dependent LULC class, resampled using nearest neighbour.
            * **erosion_usle_c** map: Cover management factor from the USLE equation [-]
            * **soil_compacted_fraction** map: The fraction of compacted or urban area
                per grid cell [-]
            * **land_water_fraction** map: The fraction of water covered area per grid
              cell [-]

        Required setup methods:

        * :py:meth:`~WflowSedimentModel.setup_basemaps`

        Parameters
        ----------
        lulc_fn : str, gpd.GeoDataFrame
            GeoDataFrame or name in data catalog / path to (vector) landuse map.

            * Required columns: 'landuse' [-]
        lulc_mapping_fn : str, Path, pd.DataFrame, None, optional
            Path to a mapping csv file from landuse in source name to parameter values
            in lulc_vars. If lulc_fn is one of {"globcover", "vito", "corine",
            "esa_worldcover", "glmnco"}, a default mapping is used and this argument
            becomes optional.
        planted_forest_fn : str, Path, gpd.GeoDataFrame, None, optional
            GeoDataFrame source with polygons of planted forests.

            * Optional variable: ["forest_type"]
        lulc_vars : list[str], optional
            List of landuse parameters to prepare.
            The names are the columns of the mapping file.
            Can be a subset of: ["landuse", "soil_compacted_fraction", "erosion_usle_c",
            "land_water_fraction"]
        lulc_res : float, int, None, optional
            Resolution of the intermediate rasterized landuse map. The unit (meter or
            degree) depends on the CRS of lulc_fn (projected or not). By default None,
            which uses the model resolution.
        all_touched : bool, optional
            If True, all pixels touched by the vector will be burned in the raster,
            by default False.
        buffer : int, optional
            Buffer around the bounding box of the vector data to ensure that all
            landuse classes are included in the rasterized map, by default 1000.
        save_raster_lulc : bool, optional
            If True, the high resolution rasterized landuse map will be saved to
            maps/landuse_raster.tif, by default False.
        planted_forest_c : float, optional
            Value of USLE C factor for planted forest, by default 0.0881.
        orchard_name : str, optional
            Name of orchard landuse class in the "forest_type" column of
            ``planted_forest_fn``, by default "Orchard".
        orchard_c : float, optional
            Value of USLE C factor for orchards, by default 0.2188.
        output_names_suffix : str, optional
            Suffix to be added to the output names to avoid having to rename all the
            columns of the mapping tables. For example if the suffix is "vito", all
            variables in lulc_vars will be renamed to "landuse_vito",
            "erosion_usle_c_vito", etc.

        See Also
        --------
        workflows.setup_lulcmaps_from_vector
        workflows.add_planted_forest_to_landuse
        """
        # Prepare all default parameters
        super().setup_lulcmaps_from_vector(
            lulc_fn=lulc_fn,
            lulc_mapping_fn=lulc_mapping_fn,
            lulc_vars=lulc_vars,
            lulc_res=lulc_res,
            all_touched=all_touched,
            buffer=buffer,
            save_raster_lulc=save_raster_lulc,
            output_names_suffix=output_names_suffix,
        )

        # If available, improve USLE C map with planted forest data
        if self._MAPS["usle_c"] in lulc_vars and planted_forest_fn is not None:
            # Read forest data
            planted_forest = self.data_catalog.get_geodataframe(
                planted_forest_fn,
                geom=self.basins,
                buffer=1,
                predicate="intersects",
                handle_nodata="IGNORE",
            )
            if planted_forest is None:
                logger.warning("No Planted forest data found within domain.")
                return
            rename_dict = {
                v: k
                for k, v in self._MAPS.items()
                if v in self.staticmaps.data.data_vars
            }
            usle_c = workflows.add_planted_forest_to_landuse(
                planted_forest,
                self.staticmaps.data.rename(rename_dict),
                planted_forest_c=planted_forest_c,
                orchard_name=orchard_name,
                orchard_c=orchard_c,
            )

            # Add to grid
            self.staticmaps.set(usle_c, name=self._MAPS["usle_c"])

    @hydromt_step
    def setup_canopymaps(
        self,
        canopy_fn: str | Path | xr.DataArray,
        output_name: str = "vegetation_height",
    ):
        """Generate sediments based canopy height maps.

        Adds model layers:

            * **vegetation_height** map: height of the vegetation canopy [m]

        Parameters
        ----------
        canopy_fn :
            Canopy height data source (DataArray).
        output_name : dict, optional
            Name of the output map. By default 'vegetation_height'.
        """
        logger.info("Preparing canopy height map.")

        # Canopy height
        dsin = self.data_catalog.get_rasterdataset(
            canopy_fn, geom=self.region, buffer=2
        )
        dsout = xr.Dataset(coords=self.staticmaps.data.raster.coords)
        ds_out = dsin.raster.reproject_like(self.staticmaps.data, method="average")
        dsout["vegetation_height"] = ds_out.astype(np.float32)
        dsout["vegetation_height"] = dsout["vegetation_height"].fillna(-9999.0)
        dsout["vegetation_height"].raster.set_nodata(-9999.0)

        # update name
        wflow_var = self._WFLOW_NAMES[self._MAPS["vegetation_height"]]
        self._update_naming({wflow_var: output_name})
        self.staticmaps.set(dsout["vegetation_height"], name=output_name)
        # update config
        self._update_config_variable_name(output_name)

    @hydromt_step
    def setup_soilmaps(
        self,
        soil_fn: str = "soilgrids",
        usle_k_method: str = "renard",
        add_aggregates: bool = True,
        output_names: dict = {
            "soil_clay__mass_fraction": "soil_clay_fraction",
            "soil_silt__mass_fraction": "soil_silt_fraction",
            "soil_sand__mass_fraction": "soil_sand_fraction",
            "soil_small_aggregates__mass_fraction": "soil_sagg_fraction",
            "soil_large_aggregates__mass_fraction": "soil_lagg_fraction",
            "soil_erosion__rainfall_soil_detachability_factor": "erosion_soil_detachability",  # noqa: E501
            "soil_erosion__usle_k_factor": "erosion_usle_k",
            "land_surface_sediment__median_diameter": "soil_sediment_d50",
            "land_surface_water_sediment__govers_transport_capacity_coefficient": "land_govers_c",  # noqa: E501
            "land_surface_water_sediment__govers_transport_capacity_exponent": "land_govers_n",  # noqa: E501
        },
    ):
        """Generate sediments based soil parameter maps.

        Sediment size distribution and addition of small and large aggregates can be
        estimated from primary particle size distribution with Foster et al. (1980).
        USLE K factor can be computed from the soil data using Renard or EPIC methods.
        Calculation of D50 and fraction of fine and very fine sand (fvfs) from
        Fooladmand et al, 2006.

        Adds model layers:

            * **soil_clay_fraction**: clay content of the topsoil [g/g]
            * **soil_silt_fraction**: silt content of the topsoil [g/g]
            * **soil_sand_fraction**: sand content of the topsoil [g/g]
            * **soil_sagg_fraction**: small aggregate content of the topsoil [g/g]
            * **soil_lagg_fraction**: large aggregate content of the topsoil [g/g]
            * **erosion_soil_detachability** map: mean detachability of the soil
                (Morgan et al., 1998) [g/J]
            * **erosion_usle_k** map: soil erodibility factor from the USLE equation
                [-]
            * **soil_sediment_d50** map: median sediment diameter of the soil [mm]
            * **land_govers_c** map: Govers factor for overland flow transport
                capacity [-]
            * **land_govers_n** map: Govers exponent for overland flow transport
                capacity [-]

        Required setup methods:

        * :py:meth:`~WflowSedimentModel.setup_basemaps`


        Parameters
        ----------
        soil_fn : {"soilgrids"}
            Name of soil data source in data_sources.yml file.
                * Required variables: ['clyppt_sl1', 'sltppt_sl1', 'oc_sl1']
        usle_k_method: {"renard", "epic"}
            Method to compute the USLE K factor, by default renard.
        add_aggregates: bool, optional
            Add small and large aggregates based on soil texture, by default True.
        output_names : dict, optional
            Dictionary with output names that will be used in the model netcdf input
            files. Users should provide the Wflow.jl variable name followed by the name
            in the netcdf file.
        """
        logger.info("Preparing soil parameter maps.")

        # Soil related maps
        if soil_fn not in ["soilgrids"]:
            logger.warning(
                f"Invalid source '{soil_fn}', skipping setup_soilmaps for sediment."
            )
            return

        self._update_naming(output_names)
        dsin = self.data_catalog.get_rasterdataset(soil_fn, geom=self.region, buffer=2)
        dsout = workflows.soilgrids_sediment(
            dsin,
            self.staticmaps.data,
            usle_k_method=usle_k_method,
            add_aggregates=add_aggregates,
        )
        rmdict = {k: self._MAPS.get(k, k) for k in dsout.data_vars}
        self.staticmaps.set(dsout.rename(rmdict))
        self._update_config_variable_name(dsout.rename(rmdict).data_vars)

    @hydromt_step
    def setup_outlets(
        self,
        river_only: bool = True,
        toml_output: str = "csv",
        gauge_toml_header: list[str] = ["suspended_solids"],
        gauge_toml_param: list[str] = [
            "river_water_sediment__suspended_mass_concentration",
        ],
    ):
        """Set the default gauge map based on basin outlets.

        If the subcatchment map is available, the catchment outlets IDs will be matching
        the subcatchment IDs. If not, then IDs from 1 to number of outlets are used.

        Can also add csv/netcdf_scalar output settings in the TOML.

        Adds model layers:

        * **outlets** map: IDs map from catchment outlets [-]
        * **outlets** geom: polygon of catchment outlets

        Required setup methods:

        * :py:meth:`~WflowSedimentModel.setup_rivers`

        Parameters
        ----------
        river_only : bool, optional
            Only derive outlet locations if they are located on a river instead of
            locations for all catchments, by default True.
        toml_output : str, optional
            One of ['csv', 'netcdf_scalar', None] to update [output.csv] or
            [output.netcdf_scalar] section of wflow toml file or do nothing. By
            default, 'csv'.
        gauge_toml_header : list, optional
            Save specific model parameters in csv section. This option defines
            the header of the csv file.
            By default saves suspended_solids (for
            river_water_sediment__suspended_mass_concentration).
        gauge_toml_param: list, optional
            Save specific model parameters in csv section. This option defines
            the wflow variable corresponding to the names in gauge_toml_header.
            By default saves river_water_sediment__suspended_mass_concentration (for
            suspended_solids).
        """
        super().setup_outlets(
            river_only=river_only,
            toml_output=toml_output,
            gauge_toml_header=gauge_toml_header,
            gauge_toml_param=gauge_toml_param,
        )

    @hydromt_step
    def setup_gauges(
        self,
        gauges_fn: str | Path | gpd.GeoDataFrame,
        index_col: str | None = None,
        snap_to_river: bool = True,
        mask: np.ndarray | None = None,
        snap_uparea: bool = False,
        max_dist: float = 10e3,
        wdw: int = 3,
        rel_error: float = 0.05,
        abs_error: float = 50.0,
        fillna: bool = False,
        derive_subcatch: bool = False,
        basename: str | None = None,
        toml_output: str | None = "csv",
        gauge_toml_header: list[str] | None = ["river_q", "suspended_solids"],
        gauge_toml_param: list[str] | None = [
            "river_water__volume_flow_rate",
            "river_water_sediment__suspended_mass_concentration",
        ],
        **kwargs,
    ):
        """Set a gauge map based on ``gauges_fn`` data.

        This function directly calls the ``setup_gauges`` function of the
        WflowBaseModel, see
        py:meth:`hydromt_wflow.wflow_base.WflowBaseModel.setup_gauges` for more details.

        The only differences are the default values for the arguments:

        - ``gauge_toml_header`` defaults to ["river_q", "suspended_solids"]
        - ``gauge_toml_param`` defaults to ["river_water__volume_flow_rate",
            "river_water_sediment__suspended_mass_concentration"]

        Required setup methods:

        * :py:meth:`~WflowSedimentModel.setup_rivers`

        See Also
        --------
        WflowBaseModel.setup_gauges
        """
        # # Add new outputcsv section in the config
        super().setup_gauges(
            gauges_fn=gauges_fn,
            index_col=index_col,
            snap_to_river=snap_to_river,
            mask=mask,
            snap_uparea=snap_uparea,
            max_dist=max_dist,
            wdw=wdw,
            rel_error=rel_error,
            abs_error=abs_error,
            fillna=fillna,
            derive_subcatch=derive_subcatch,
            basename=basename,
            toml_output=toml_output,
            gauge_toml_header=gauge_toml_header,
            gauge_toml_param=gauge_toml_param,
            **kwargs,
        )

    @hydromt_step
    def clip(
        self,
        region: dict,
        inverse_clip: bool = False,
        clip_forcing: bool = True,
        clip_states: bool = True,
        crs: int = 4326,
        **kwargs,
    ):
        """Clip model to region.

        First the staticmaps are clipped to the region.
        Then the staticgeoms are re-generated to match the new grid for basins and
        rivers and clipped for the others.
        Finally the forcing and states are clipped to the new grid extent.

        Parameters
        ----------
        region : dict
            See :meth:`models.wflow_base.WflowBaseModel.setup_basemaps`
        inverse_clip: bool, optional
            Flag to perform "inverse clipping": removing an upstream part of the model
            instead of the subbasin itself, by default False
        clip_forcing: bool, optional
            Flag to clip the forcing to the new grid extent, by default True
        clip_states: bool, optional
            Flag to clip the states to the new grid extent, by default True
        crs: int, optional
            Default crs of the grid to clip.
        **kwargs: dict
            Additional keyword arguments passed to
            :py:meth:`~hydromt.raster.Raster.clip_geom`
        """
        # Reservoir maps that will be removed if no reservoirs after clipping
        # key: staticmaps name,  value: wflow input variable name
        reservoir_maps = [
            self._MAPS["reservoir_area_id"],
            self._MAPS["reservoir_outlet_id"],
            self._MAPS["reservoir_area"],
            self._MAPS["reservoir_trapping_efficiency"],
        ]
        reservoir_maps = {k: self._WFLOW_NAMES.get(k, None) for k in reservoir_maps}

        super().clip(
            region,
            inverse_clip=inverse_clip,
            clip_forcing=clip_forcing,
            clip_states=clip_states,
            reservoir_maps=reservoir_maps,
            crs=crs,
            **kwargs,
        )

    @hydromt_step
    def upgrade_to_v1_wflow(
        self,
        soil_fn: str = "soilgrids",
        usle_k_method: str = "renard",
        strord_name: str = "wflow_streamorder",
    ):
        """
        Upgrade the model to wflow v1 format.

        The function reads a TOML from wflow v0x and converts it to wflow v1x format.
        The other components stay the same.

        A few variables that used to be computed within Wflow.jl are now moved to
        HydroMT to allow more flexibility for the users to update if they do get local
        data or calibrate some of the parameters specifically. For this, the
        ``setup_soilmaps`` and ``setup_riverbedsed`` functions are called again.

        Lakes and reservoirs have also been merged into one structure and parameters in
        the resulted staticmaps will be combined.

        This function should be followed by ``write_config`` to write the upgraded TOML
        file and by ``write_grid`` to write the upgraded static netcdf input file.

        Parameters
        ----------
        soil_fn : str, optional
            soil_fn argument of setup_soilmaps method.
        usle_k_method : str, optional
            usle_k_method argument of setup_soilmaps method.
        strord_name : str, optional
            strord_name argument of setup_riverbedsed method.
        """
        config_v0 = self.config.data.copy()
        config_out = convert_to_wflow_v1_sediment(self.config.data)

        # Update the config
        with open(utils.DATADIR / "default_config_headers.toml", "rb") as file:
            self.config._data = tomllib.load(file)
        for option in config_out:
            self.config.set(option, config_out[option])

        # Rerun setup_soilmaps
        self.setup_soilmaps(
            soil_fn=soil_fn,
            usle_k_method=usle_k_method,
            add_aggregates=True,
        )

        # Rerun setup_riverbedsed
        self.setup_riverbedsed(bedsed_mapping_fn=None, strord_name=strord_name)

        # Merge lakes and reservoirs layers
        ds_res, vars_to_remove, config_opt = convert_reservoirs_to_wflow_v1_sediment(
            self.staticmaps.data, config_v0
        )
        if ds_res is not None:
            # Remove older maps from grid
            self.staticmaps.drop_vars(vars_to_remove)
            # Add new reservoir maps to grid
            self.staticmaps.set(ds_res)
            # Update the config with the new names
            for option in config_opt:
                self.config.set(option, config_opt[option])



import logging
from collections.abc import Callable
from pathlib import Path
from unittest.mock import MagicMock

import pytest
import xarray as xr

from hydromt_wflow.components import WflowOutputScalarComponent
from hydromt_wflow.wflow_sbm import WflowSbmModel


@pytest.fixture
def mock_model(mock_model_factory) -> WflowSbmModel:
    """Fixture to create a mock WflowSbmModel."""
    return mock_model_factory(mode="w")


# Define the side effect function
def get_value_side_effect(key, fallback=None):
    if key == "dir_output":
        return "run_results"
    elif key == "output.netcdf_scalar.path":
        return "output_scalar.nc"
    return fallback


def test_wflow_output_scalar_component_init(mock_model: WflowSbmModel):
    # Setup the component
    component = WflowOutputScalarComponent(mock_model)

    # Assert that the internal data is None
    assert component._data is None

    # When asking for data property, it should return a xarray.Dataset
    assert isinstance(component.data, xr.Dataset)
    assert isinstance(component._data, xr.Dataset)  # Same for internal
    assert len(component.data) == 0


def test_wflow_output_scalar_component_set(
    mock_model: WflowSbmModel, forcing_layer: xr.DataArray
):
    # Setup the component
    component = WflowOutputScalarComponent(mock_model)

    # Assert that the internal data is None, initializing will happen in set
    assert component._data is None

    # Set an entry
    # Copy else the name gets altered for the next test?
    component.set(forcing_layer.copy(), name="test_layer")

    # Assert the content
    assert isinstance(component._data, xr.Dataset)
    assert "test_layer" in component.data
    assert len(component.data) == 1


def test_wflow_output_scalar_component_set_alt(
    mock_model: WflowSbmModel, forcing_layer: xr.DataArray
):
    # Setup the component
    component = WflowOutputScalarComponent(mock_model)

    # Assert that the internal data is None, initializing will happen in set
    assert component._data is None

    # Set an entry
    # Copy else the name gets altered for the next test?
    forcing_layer.name = "precip"
    component.set(forcing_layer.copy())

    # Assert the content
    assert "precip" in component.data


def test_wflow_output_scalar_component_set_errors(
    mock_model: WflowSbmModel, forcing_layer: xr.DataArray
):
    # Setup the component
    component = WflowOutputScalarComponent(mock_model)

    # Missing name
    forcing_layer.name = None
    with pytest.raises(ValueError, match="Name required for DataArray."):
        component.set(forcing_layer)

    # Try with np.ndarray
    array = forcing_layer.values
    with pytest.raises(TypeError, match="cannot set data of type ndarray"):
        component.set(array)


def test_wflow_output_scalar_component_read(
    mock_model_factory: Callable[[Path, str], WflowSbmModel],
    model_subbasin_cached: Path,
):
    # Set it to read mode
    mock_model = mock_model_factory(path=model_subbasin_cached, mode="r")

    # Setup the component
    component = WflowOutputScalarComponent(mock_model)

    # Assert its data currently none
    assert component._data is None

    # Read the data
    type(component.model.config).get_value = MagicMock(
        side_effect=get_value_side_effect
    )
    component.read()

    # Assert the read data
    assert isinstance(component.data, xr.Dataset)
    assert len(component.data) == 2
    assert "temp_coord" in component.data


def test_wflow_output_scalar_component_read_init(
    mock_model_factory: Callable[[Path, str], WflowSbmModel],
    model_subbasin_cached: Path,
):
    # Set it to read mode
    mock_model = mock_model_factory(path=model_subbasin_cached, mode="r")

    # Setup the component
    component = WflowOutputScalarComponent(mock_model)
    type(component.model.config).get_value = MagicMock(
        side_effect=get_value_side_effect
    )
    assert component._data is None  # Assert no data or structure yet

    # Read at init
    assert len(component.data) == 2
    assert "temp_coord" in component.data


def test_wflow_output_scalar_component_write(
    mock_model: WflowSbmModel,
    grid_dummy_data: xr.DataArray,
    caplog: pytest.LogCaptureFixture,
):
    # Setup the component
    component = WflowOutputScalarComponent(mock_model)
    component._data = grid_dummy_data.to_dataset(name="test_layer")

    # Write to a file
    # Check the logger message when calling component.write()
    with caplog.at_level(logging.INFO):
        component.write()
    assert "netcdf_scalar is an output of Wflow and will not be written." in caplog.text


def test_wflow_output_scalar_component_equal(
    mock_model: WflowSbmModel,
    grid_dummy_data: xr.DataArray,
):
    # Setup the components
    component = WflowOutputScalarComponent(mock_model)
    component2 = WflowOutputScalarComponent(mock_model)

    # Update them manually
    component._data = grid_dummy_data.to_dataset()
    component2._data = grid_dummy_data.to_dataset()

    # Assert these are equal
    eq, errors = component.test_equal(component2)
    assert eq is True

    # Update component2 to make them not equal
    component2._data["test_layer"] = grid_dummy_data + 1  # Modify data

    # Assert unequal
    eq, errors = component.test_equal(component2)
    assert eq is False
    assert "Data variables only on the right object" in errors["data"]
    assert "test_layer" in errors["data"]



"""Data for examples and testing of HydroMT-wflow."""

import json
import logging
from pathlib import Path

import pooch

__all__ = ["fetch_data"]

logger = logging.getLogger(f"hydromt.{__name__}")

PROCESSORS = {
    "tar.gz": pooch.Untar,
    "zip": pooch.Unzip,
}


def unpack_processor(
    suffix: str,
    extract_dir: str = "./",
):
    """Select the right processor for unpacking."""
    if suffix not in PROCESSORS:
        return None
    processor = PROCESSORS[suffix](members=None, extract_dir=extract_dir)
    return processor


def fetch_data(
    data: str,
    output_dir: Path | str | None = None,
    sub_dir: bool = True,
) -> Path:
    """Fetch data by simply calling the function.

    Parameters
    ----------
    data : str
        The data to fetch.
    output_dir : Path | str | None
        The output directory to store the data.
        If None, the data will be stored in ~/.cache/hydromt_wflow/<data>
    sub_dir : bool
        Whether to place the fetched data in a sub directory of the same name.
        I.e. if  the (tarred) dataset is named 'custom-data' a directory named
        'custom-data' is created in which the data are placed. By default True

    Returns
    -------
    Path
        The output directory where the data is stored
    """
    # Open the registry
    # update the base URL and registry with new versions of the data
    # use create_artifact.py script to create the build-data/ test-data archives
    with open(Path(__file__).parent / "registry.json", "r") as f:
        database = json.load(f)
        base_url: str = database["url"]
        registry: dict[str:str] = database["data"]
    cache_dir = Path("~", ".cache", "hydromt_wflow").expanduser()

    cache_dir.mkdir(parents=True, exist_ok=True)
    if output_dir is None:
        output_dir = cache_dir
    else:
        output_dir = Path(output_dir)
        output_dir.mkdir(parents=True, exist_ok=True)

    # Quick check whether the data can be found
    choices_raw = list(registry.keys())
    choices = [item.split(".", 1)[0] for item in choices_raw]
    if data not in choices:
        raise ValueError(f"Choose one of the following: {choices}")
    idx = choices.index(data)

    # Setup Pooch
    retriever = pooch.create(
        path=cache_dir,  # store archive to cache
        base_url=base_url,
        registry=registry,
    )

    # Set the way of unpacking it
    suffix = choices_raw[idx].split(".", 1)[1]
    extract_dir = output_dir
    if sub_dir:
        extract_dir = Path(extract_dir, data)
    processor = unpack_processor(suffix, extract_dir=extract_dir)
    # Retrieve the data
    retriever.fetch(choices_raw[idx], processor=processor)

    return extract_dir



import logging
from collections.abc import Callable
from pathlib import Path
from unittest.mock import MagicMock, PropertyMock

import numpy as np
import pytest
import xarray as xr
from hydromt.model import ModelRoot

from hydromt_wflow.components import WflowStaticmapsComponent
from hydromt_wflow.wflow_sbm import WflowSbmModel


@pytest.fixture
def mock_model(mock_model_factory) -> MagicMock:
    """Create a mock model with a root."""
    return mock_model_factory(mode="w")


def test_wflow_staticmaps_component_init(mock_model: MagicMock):
    # Setup the component
    component = WflowStaticmapsComponent(mock_model)

    # Assert that the internal data is None
    assert component._data is None

    # When asking for data property, it should return a xr.Dataset
    assert isinstance(component.data, xr.Dataset)
    assert isinstance(component._data, xr.Dataset)  # Same for internal
    assert len(component.data) == 0  # Zero data vars


def test_wflow_staticmaps_component_set(
    mock_model: MagicMock,
    static_layer: xr.DataArray,
):
    # Setup the component
    component = WflowStaticmapsComponent(mock_model)

    # Set the data in the component
    component.set(static_layer, "layer1")
    # Set with name included in the dataarray
    static_layer.name = "layer2"
    component.set(static_layer * 2)

    # Assert the data
    assert "layer1" in component.data.data_vars
    assert "layer2" in component.data.data_vars
    assert list(component.data.dims) == ["lat", "lon"]


def test_wflow_staticmaps_component_set_cyclic(
    mock_model: MagicMock,
    cyclic_layer: xr.DataArray,
    cyclic_layer_large: xr.DataArray,
):
    # Setup the component
    component = WflowStaticmapsComponent(mock_model)

    # Set the data in the component
    component.set(cyclic_layer, "layer1")

    # Assert the current state
    assert "layer1" in component.data.data_vars
    assert "time" in component.data.dims
    assert len(component.data.time) == 12

    # Add another but with a different valid time dimension
    component.set(cyclic_layer_large, "layer2")

    # Assert the data and dimensions
    assert "layer2" in component.data.data_vars
    assert "time_365" in component.data.dims
    assert len(component.data.time_365) == 365
    assert "time_365" in component.data["layer2"].coords

    # Set the data in the component
    component.set(cyclic_layer * 2, "layer3")

    # Assert the current state
    assert "layer3" in component.data.data_vars
    assert "time" in component.data.dims  # Still there, not changed


def test_wflow_staticmaps_component_set_mask(
    mock_model: MagicMock,
    static_layer: xr.DataArray,
    mask_layer: xr.DataArray,
):
    # Setup the component
    component = WflowStaticmapsComponent(mock_model)

    # Set the mask layer in the dataset and use that as the mask in the set method
    component.set(mask_layer, name="mask")
    mock_model._MAPS = {"basins": "mask"}  # Needed as internally references like this
    component.set(static_layer, name="layer2")

    # Assert that layer2 has been masked
    np.testing.assert_array_equal(
        component.data["layer2"],
        np.array([[1, -9999], [1, 1]]),
    )


def test_wflow_staticmaps_component_set_errors(
    mock_model: MagicMock,
    cyclic_layer: xr.DataArray,
):
    # Setup the component
    component = WflowStaticmapsComponent(mock_model)

    # No naming given either via the dataarray or argument
    with pytest.raises(ValueError, match="Unable to set DataArray data without a name"):
        component.set(cyclic_layer)

    # Wrong data input type, list is of course not accepted, only xarray stuff
    with pytest.raises(
        ValueError,
        match="Cannot set data of type list",
    ):
        component.set([2, 2])

    # Wrong length of cyclic data, either monthly or daily
    with pytest.raises(
        ValueError,
        match=r"Length of cyclic dataset \(11\) is not supported",
    ):
        # Drop 1 so the length becomes 11, which is invalid
        component.set(
            cyclic_layer.where(cyclic_layer.time != 1, drop=True),
            name="layer1",
        )


def test_wflow_staticmaps_component_set_warnings(
    caplog: pytest.LogCaptureFixture,
    mock_model: MagicMock,
    static_layer: xr.DataArray,
    cyclic_layer: xr.DataArray,
    cyclic_layer_large: xr.DataArray,
):
    caplog.set_level(logging.INFO)
    # Setup the component
    component = WflowStaticmapsComponent(mock_model)

    # Set the data
    component.set(static_layer, "layer1")
    component.set(cyclic_layer.rename(time="layer"), "layer2")

    # First warning when replacing a layer that is already present
    component.set(static_layer * 2, "layer1")
    # Assert the logging message
    assert "Replacing grid map: layer1" in caplog.text

    # Second warning when setting a layer with a 'layer' dimension that differs
    # from the layer dimension that is already in there
    component.set(cyclic_layer_large.rename(time="layer"), "layer3")
    # Assert the logging message
    assert (
        "Replacing 'layer' coordinate, dropping variables \
(['layer2']) associated with old coordinate"
        in caplog.text
    )


def test_wflow_staticmaps_component_update_names(
    mock_model: MagicMock,
    static_layer: xr.DataArray,
):
    # Setup the component
    component = WflowStaticmapsComponent(mock_model)

    # Set data like a dummy
    component._data = static_layer.to_dataset(name="layer1")
    component.data["layer2"] = static_layer * 2
    component.data["layer3"] = static_layer * 3

    # Rename some variables
    component.update_names(layer1="foo", layer2="bar")

    # Assert it succeded
    assert "foo" in component.data.data_vars
    assert "bar" in component.data.data_vars


def test_wflow_staticmaps_component_update_names_warming(
    caplog: pytest.LogCaptureFixture,
    mock_model: MagicMock,
    static_layer: xr.DataArray,
):
    caplog.set_level(logging.INFO)
    # Setup the component
    component = WflowStaticmapsComponent(mock_model)

    # Set data like a dummy
    component._data = static_layer.to_dataset(name="layer1")

    # Rename but faulty
    component.update_names(layer2="foo")
    # Assert the warning message
    assert "Could not rename ['layer2'], not found in data" in caplog.text


def test_wflow_staticmaps_component_read(
    mock_model_factory: Callable[[Path, str], WflowSbmModel],
    model_subbasin_cached: Path,
):
    # Set the root to model dir in read mode
    mock_model = mock_model_factory(path=model_subbasin_cached, mode="r")

    # Setup the component
    component = WflowStaticmapsComponent(mock_model)
    # Assert the internal data is None
    assert component._data is None

    # Read the data
    type(component.model.config).get_value = MagicMock(return_value="")
    component.read()

    # Assert the data
    assert isinstance(component.data, xr.Dataset)
    assert "meta_landuse" in component.data.data_vars
    assert len(component.data) == 69  # 69 layers


def test_wflow_staticmaps_component_read_empty(
    tmp_path: Path,
    mock_model: MagicMock,
):
    # Set the root to model dir in read mode
    type(mock_model).root = PropertyMock(
        side_effect=lambda: ModelRoot(tmp_path, mode="r"),
    )

    # Setup the component
    component = WflowStaticmapsComponent(mock_model)
    # Assert the internal data is None
    assert component._data is None

    # Read the data
    type(component.model.config).get_value = MagicMock(return_value="")
    component.read()
    # Assert that its an emptry Dataset
    assert isinstance(component.data, xr.Dataset)
    assert len(component.data) == 0


def test_wflow_staticmaps_component_write(
    mock_model: MagicMock,
    static_layer: xr.DataArray,
):
    # Setup the component
    component = WflowStaticmapsComponent(mock_model)
    mock_model.components = {"staticmaps": component}

    # Set the data like a dummy
    component._data = static_layer.to_dataset(name="layer1")

    # Write the data
    type(component.model.config).get_value = MagicMock(return_value="")  # dir_input
    component.write()

    # Assert the output (and config)
    assert Path(component.root.path, component._filename).is_file()



"""Some utilities from the Wflow plugin."""

import functools
import logging
import os
from pathlib import Path
from typing import Any, Callable, Union

import geopandas as gpd
import numpy as np
import xarray as xr
from hydromt.gis import GeoDataArray
from hydromt.model.processes.grid import grid_from_constant
from hydromt.readers import open_timeseries_from_table

logger = logging.getLogger(f"hydromt.{__name__}")

DATADIR = Path(__file__).parent / "data"

__all__ = [
    "get_config",
    "set_config",
    "get_grid_from_config",
    "read_csv_output",
]


def get_config(
    config: dict,
    key: str,
    root: Path | None = None,
    fallback: Any | None = None,
    abs_path: bool = False,
):
    """
    Retrieve a nested configuration value from a dictionary.

    Keys can be specified using dot-separated paths (e.g. ``"a.b.c"``) to
    access nested dictionaries. Optionally, the function can resolve relative
    paths against a given model root.

    Parameters
    ----------
    config : dict
        The configuration dictionary to read from.
    key : str
        Dotted key path indicating the value to retrieve, e.g. ``"a.b.c"``.
    root : Path, optional
        Path to the model root. Used when ``abs_path=True`` to resolve
        relative paths. Default is ``None``.
    fallback : Any, optional
        Value to return if the key path does not exist. Default is ``None``.
    abs_path : bool, optional
        If ``True`` and the retrieved value is a relative path, return its
        absolute version relative to ``root``. Default is ``False``.

    Returns
    -------
    Any
        The value corresponding to the given key path. Returns ``fallback``
        if the key does not exist.

    Examples
    --------
    .. code-block:: python

       >>> from pathlib import Path
       >>> from hydromt_wflow.utils import get_config
       >>> cfg = {
       ...     "model": {
       ...         "params": {"scale": 2.5},
       ...         "paths": {"data": "input/data.csv"},
       ...     },
       ...     "verbose": True,
       ... }

       # Retrieve a simple value
       >>> get_config(cfg, "verbose")
       True

       # Retrieve a nested value
       >>> get_config(cfg, "model.paths.data")
       'input/data.csv'

       # Retrieve as absolute path relative to a model root
       >>> root = Path("/home/user/wflow_model")
       >>> get_config(cfg, "model.paths.data", root=root, abs_path=True)
       Path('/home/user/wflow_model/input/data.csv')

       # Retrieve a non-existent key with fallback
       >>> get_config(cfg, "model.output.folder", fallback="output")
       'output'
    """
    parts = key.split(".")
    num_parts = len(parts)
    current = config
    value = fallback
    for i, part in enumerate(parts):
        if i < num_parts - 1:
            current = current.get(part, {})
        else:
            value = current.get(part, fallback)

    if abs_path and isinstance(value, (str, Path)):
        value = Path(value)
        if not value.is_absolute():
            if root is None:
                raise ValueError(
                    "root path is required to get absolute path from relative path"
                )
            value = Path(os.path.abspath(os.path.join(root, value)))

    return value


def set_config(config: dict, key: str, value: Any):
    """
    Update a nested configuration dictionary using a dotted key path.

    This function modifies a configuration dictionary in-place, allowing
    assignment to nested keys using dot-separated paths (e.g. ``"a.b.c"``).

    Parameters
    ----------
    config : dict
        The configuration dictionary to update.
    key : str
        The dotted key path identifying the target value to set.
    value : Any
        The new value to assign at the specified key path.

    Returns
    -------
    dict
        The updated configuration dictionary.

    Examples
    --------
    .. code-block:: python

       >>> from hydromt_wflow.utils import set_config
       >>> cfg = {"model": {"params": {"scale": 1.0, "offset": 0.0}}}
       >>> set_config(cfg, "model.params.scale", 2.5)
       {'model': {'params': {'scale': 2.5, 'offset': 0.0}}}

       You can also create nested keys if they do not exist:

       >>> set_config(cfg, "model.runtime.threads", 8)
       {'model': {'params': {'scale': 2.5, 'offset': 0.0},
                  'runtime': {'threads': 8}}}

       Updating a top-level key works the same way:

       >>> set_config(cfg, "description", "Wflow test run")
       {'model': {'params': {'scale': 2.5, 'offset': 0.0},
                  'runtime': {'threads': 8}},
        'description': 'Wflow test run'}
    """
    if not isinstance(key, str):
        raise TypeError("key must be string")
    keys = key.split(".")
    functools.reduce(lambda d, k: d.setdefault(k, {}), keys[:-1], config)[keys[-1]] = (
        value
    )


def read_csv_output(
    fn: Path | str, config: dict, maps: xr.Dataset
) -> dict[str, GeoDataArray]:
    """Read wflow output csv timeseries and parse to dictionary.

    Parses the wflow csv output file into different ``hydromt.GeoDataArrays``, one per
    column (csv section and csv.column sections of the TOML). The xy coordinates are the
    coordinates of the station or of the representative point of the subcatch/area. The
    variable name in the ``GeoDataArray`` corresponds to the csv header attribute or
    header_map when available.

    Parameters
    ----------
    fn: str
        Path to the wflow csv output file.
    config: dict
        wflow.toml configuration.
    maps: xr.Dataset
        wflow staticmaps.nc dataset

    Returns
    -------
    csv_dict: dict
        Dictionary of hydromt.GeoDataArrays for the different csv.column section \
of the config.
    """
    # Count items by csv.column
    count = 1
    csv_dict = {}
    # Loop over csv.column
    for col in config["output"]["csv"].get("column"):
        header = col["header"]
        logger.debug(f"Reading csv column '{header}'")
        # Column based on map
        if "map" in col.keys():
            # Read the corresponding map and derive the different locations
            # The centroid of the geometry is used as coordinates for the timeseries
            map_name = config["input"].get(f"{col['map']}")
            if map_name not in maps:
                logger.warning(
                    f"Map '{map_name}' not found in staticmaps. Skip reading."
                )
                return {}
            da = maps[map_name]
            gdf = da.raster.vectorize()
            gdf.geometry = gdf.geometry.representative_point()
            gdf.index = gdf.value.astype(da.dtype)
            gdf.index.name = "index"
            # Read the timeseries
            usecols = [0]
            usecols = np.append(usecols, np.arange(count, count + len(gdf.index)))
            count += len(gdf.index)
            da_ts = open_timeseries_from_table(
                fn, name=f"{header}_{col['map']}", usecols=usecols
            )
            da = GeoDataArray.from_gdf(gdf, da_ts, index_dim="index")
        # Column based on xy coordinates or reducer for the full model domain domain
        else:
            # Read the timeseries
            usecols = [0]
            usecols = np.append(usecols, np.arange(count, count + 1))
            count += 1
            try:
                da_ts = open_timeseries_from_table(fn, name=header, usecols=usecols)
            except Exception:
                colnames = ["time", "0"]
                da_ts = open_timeseries_from_table(
                    fn,
                    name=header,
                    usecols=usecols,
                    header=0,
                    names=colnames,
                )
            # Add point coordinates
            # Column based on xy coordinates
            if "coordinate" in col.keys():
                scoords = {
                    "x": xr.IndexVariable("index", [col["coordinate"]["x"]]),
                    "y": xr.IndexVariable("index", [col["coordinate"]["y"]]),
                }
            # Column based on index
            elif "index" in col.keys():
                # x and y index, works on the full 2D grid
                if isinstance(col["index"], dict):
                    # index in julia starts at 1
                    # coordinates are always ascending
                    xi = maps.raster.xcoords.values[col["index"]["x"] - 1]
                    yi = np.sort(maps.raster.ycoords.values)[col["index"]["y"] - 1]
                    scoords = {
                        "x": xr.IndexVariable("index", [xi]),
                        "y": xr.IndexVariable("index", [yi]),
                    }
                # index of the full array
                else:
                    # Create grid with full 2D Julia indices
                    # Dimensions are ascending and ordered as (x,y,layer,time)
                    # Indices are created before ordering for compatibility with
                    # raster.idx_to_xy
                    full_index = maps[
                        f"{config['input'].get('subbasin_location__count')}"
                    ].copy()
                    _, res_y = full_index.raster.res
                    if res_y < 0:
                        full_index = full_index.reindex(
                            {
                                full_index.raster.y_dim: full_index[
                                    full_index.raster.y_dim
                                ][::-1]
                            }
                        )
                    data = np.arange(0, np.size(full_index)).reshape(
                        np.size(full_index, 0), np.size(full_index, 1)
                    )
                    full_index[:, :] = data
                    full_index = full_index.transpose(
                        full_index.raster.x_dim, full_index.raster.y_dim
                    )
                    # Index depends on the struct
                    if "reservoir" in col["parameter"]:
                        mask = maps[
                            f"{config['input'].get('reservoir_location__count')}"
                        ].copy()
                    elif "river" in col["parameter"]:
                        mask = maps[
                            f"{config['input'].get('river_location__mask')}"
                        ].copy()
                    # Else all the rest should be for the whole subcatchment
                    else:
                        mask = maps[
                            f"{config['input'].get('subbasin_location__count')}"
                        ].copy()
                    # Rearrange the mask
                    res_x, res_y = mask.raster.res
                    if res_y < 0:
                        mask = mask.reindex(
                            {mask.raster.y_dim: mask[mask.raster.y_dim][::-1]}
                        )
                    mask = mask.transpose(mask.raster.x_dim, mask.raster.y_dim)
                    # Filter and reduce full_index based on mask
                    full_index = full_index.where(mask != mask.raster.nodata, 0)
                    full_index.attrs.update(_FillValue=0)

                    mask_index = full_index.values.flatten()
                    mask_index = mask_index[mask_index != 0]
                    # idx corresponding to the wflow index
                    idx = mask_index[col["index"] - 1]
                    # Reorder full_index as (y,x) to use raster.idx_to_xy method
                    xi, yi = full_index.transpose(
                        full_index.raster.y_dim, full_index.raster.x_dim
                    ).raster.idx_to_xy(idx)
                    scoords = {
                        "x": xr.IndexVariable("index", xi),
                        "y": xr.IndexVariable("index", yi),
                    }
            # Based on model bbox center for column based on reducer for
            # the full model domain
            else:
                xmin, ymin, xmax, ymax = maps.raster.bounds
                scoords = {
                    "x": xr.IndexVariable("index", [(xmax + xmin) / 2]),
                    "y": xr.IndexVariable("index", [(ymax + ymin) / 2]),
                }
            da = da_ts.assign_coords(scoords)

        csv_dict[f"{da.name}"] = da

    return csv_dict


def get_grid_from_config(
    var_name: str,
    config: dict = {},
    grid: xr.Dataset | None = None,
    root: Path | None = None,
    abs_path: bool = False,
    nodata: int | float = -9999,
    mask_name: str | None = None,
) -> xr.DataArray:
    """
    Get actual grid values from config including scale and offset.

    Calls get_config and applies value, scale and offset if available.

    Parameters
    ----------
    var_name : string
        Wflow variable name to get from config. Input, static and cyclic will be
        deduced. Eg 'land_surface__slope'.
    config : dict
        config dict to get the values from.
    grid : xr.Dataset
        grid dataset to get the values from.
    abs_path: bool, optional
        If True return the absolute path relative to the model root,
        by default False.
    nodata: int or float, optional
        nodata value to use for the DataArray, by default -9999. Used only if the
        variable in config is described using value only.
    mask_name: str, optional
        Name of the mask variable in grid. Used only if the variable in config is
        described using value only.

    Returns
    -------
    da : xr.DataArray
        DataArray with actual grid values

    See Also
    --------
    get_config
    hydromt.workflows.grid.grid_from_constant
    """
    # get config value
    # try with input only
    var_name = get_wflow_var_fullname(var_name, config)
    var = get_config(
        key=var_name,
        config=config,
        fallback=None,
        root=root,
        abs_path=abs_path,
    )
    if var is None:
        # try with input.static.var.value
        var = config["input"]["static"].get(f"{var_name}.value", None)
    if var is None:
        raise ValueError(f"variable {var_name} not found in config.")

    # direct map in grid
    if isinstance(var, str):
        if var in grid:
            da = grid[var]
        else:
            raise ValueError(f"grid variable {var} not found in staticmaps.")

    else:  # dict type
        # constant value in config
        if isinstance(var, (int, float)) or "value" in var:
            value = var if isinstance(var, (int, float)) else var["value"]
            da = grid_from_constant(
                grid,
                constant=value,
                name=var_name,
                nodata=nodata,
                mask_name=mask_name,
            )

        # else scale and offset
        else:
            var_name = get_config(key="netcdf_variable_name", config=var)
            scale = var.get("scale", 1.0)
            offset = var.get("offset", 0.0)
            # apply scale and offset
            if var_name not in grid:
                raise ValueError(f"grid variable {var_name} not found in staticmaps.")
            da = grid[var_name] * scale + offset

    return da


def get_wflow_var_fullname(input_var: str, config: dict) -> str:
    """Get the full variable name for a Wflow variable."""
    # Check if the variable is in the input section
    if get_config(config, f"input.{input_var}") is not None:
        return f"input.{input_var}"
    # Check if the variable is in the static section
    if get_config(config, f"input.static.{input_var}") is not None:
        return f"input.static.{input_var}"
    # Check if the variable is in the cyclic section
    if get_config(config, f"input.cyclic.{input_var}") is not None:
        return f"input.cyclic.{input_var}"
    # Check if the variable is in the forcing section
    if get_config(config, f"input.forcing.{input_var}") is not None:
        return f"input.forcing.{input_var}"
    # If not found, return the original variable name
    return input_var


def _mask_data_array(data_array: xr.DataArray, mask: xr.DataArray) -> xr.DataArray:
    """Mask the data array based on the mask."""
    # If the data is boolean, we set it to False where the mask is False
    if data_array.dtype == np.bool:
        return data_array.where(mask, False)
    # Otherwise we set it to nodata where the mask is False
    else:
        return data_array.where(mask, data_array.raster.nodata)


def mask_raster_from_layer(
    data: xr.Dataset | xr.DataArray, mask: xr.DataArray
) -> xr.Dataset | xr.DataArray:
    """Mask the data in the supplied grid based on the value in one of the layers.

        This for example can be used to mask a grid based on subcatchment data.
        All data in the rest of the data variables in the dataset will be set to
        the `raster.nodata` value except for boolean variables which will be set to
        False. If the supplied grid is not an xarray dataset, or does not contain
        the correct layer, it will be returned unaltered.

        The layer supplied can be either boolean or numeric. Array elements where the
        layer is larger than 0 (after type conversion) will be masked.

    Parameters
    ----------
        data (xr.Dataset, xr.DataArray):
            The grid data containing the data that should be masked
        mask (xr.DataArray):
            mask that the data will be masked to. Values can be boolean or numeric.
            Places where this layer is different than the raster nodata will be
            masked in the other data.

    Returns
    -------
        xr.Dataset, xr.DataArray: The grid with all of the data variables masked.
    """
    # Skip masking if different grid
    if data.sizes != mask.sizes:
        logger.warning("Skipping masking due to different grid sizes.")
        return data

    mask = mask != mask.raster.nodata

    if isinstance(data, xr.DataArray):
        data = _mask_data_array(data, mask)
    else:
        for var in data.data_vars:
            data[var] = _mask_data_array(data[var], mask)

    return data


def planar_operation_in_utm(
    gdf: gpd.GeoDataFrame,
    operation: Callable[[gpd.GeoSeries], Union[gpd.GeoSeries, gpd.GeoDataFrame]],
) -> Union[gpd.GeoSeries, gpd.GeoDataFrame]:
    """
    Apply a planar geometric operation on a GeoDataFrame's geometry.

    To ensure the operation is performed after converting to an appropriate UTM CRS,
    then reprojected the result back to the original CRS.

    Parameters
    ----------
        gdf (GeoDataFrame): Input GeoDataFrame with a defined CRS.
        operation (Callable): A function that operates on gdf.geometry, such as:
                              lambda geom: geom.centroid, geom.buffer(...), etc.

    Returns
    -------
        GeoSeries or GeoDataFrame: Result of the operation, reprojected to the original
        CRS.

    Examples
    --------
        >>> import geopandas as gpd
        >>> from shapely.geometry import (
        ...     Point,
        ... )
        >>> gdf = gpd.GeoDataFrame(
        ...     {
        ...         "geometry": [
        ...             Point(
        ...                 1, 2
        ...             )
        ...         ]
        ...     },
        ...     crs="EPSG:4326",
        ... )
        >>> centroid = planar_operation_in_utm(
        ...     gdf,
        ...     lambda geom: geom.centroid,
        ... )

    """
    if gdf.crs is None:
        raise ValueError("Input GeoDataFrame must have a defined CRS.")

    original_crs = gdf.crs
    utm_crs = gdf.estimate_utm_crs()

    gdf_projected = gdf.to_crs(utm_crs)
    result = operation(gdf_projected.geometry)

    if isinstance(result, gpd.GeoSeries):
        return result.set_crs(utm_crs).to_crs(original_crs)
    elif isinstance(result, gpd.GeoDataFrame):
        return result.to_crs(original_crs)
    else:
        raise TypeError("Operation must return a GeoSeries or GeoDataFrame.")



import logging

import geopandas as gpd
import numpy as np
import pytest
import xarray as xr
from hydromt import DataCatalog
from pytest_mock import MockerFixture
from shapely.geometry import box

from hydromt_wflow.workflows.basemaps import parse_region


class TestParseRegion:
    """Tests for the parse_region method."""

    @pytest.fixture
    def mock_datacatalog(
        self, mock_rasterdataset: xr.Dataset, mocker: MockerFixture
    ) -> DataCatalog:
        """Mock DataCatalog, a raster dataset with a geographic CRS and resolution."""
        mock_data_catalog = mocker.create_autospec(DataCatalog, instance=True)
        mock_data_catalog.get_rasterdataset.return_value = mock_rasterdataset
        mock_data_catalog.get_source.return_value = "mock_basin_index"
        return mock_data_catalog

    @pytest.fixture
    def mock_xy(self) -> tuple[float, float]:
        """Mock coordinates for testing."""
        return (12.2051, 45.8331)

    @pytest.fixture
    def mock_region(self, mock_xy) -> dict:
        """Mock a region dictionary."""
        x, y = mock_xy
        return {
            "basin": [x, y],
            "strord": 4,
            "bounds": [x - 1, y - 1, x + 1, y + 1],
        }

    @pytest.fixture
    def mock_get_basin_geometry(self, mock_geometry, mock_xy, mocker):
        """Mock the get_basin_geometry function."""
        mock_get_basin_geometry = mocker.patch(
            "hydromt_wflow.workflows.basemaps.get_basin_geometry"
        )
        x, y = mock_xy
        test_xy = np.array([[x, y]])
        mock_get_basin_geometry.return_value = (mock_geometry, test_xy)
        return mock_get_basin_geometry

    @pytest.fixture
    def mock_geometry(self, mock_xy) -> gpd.GeoDataFrame:
        """Mock a geometry GeoDataFrame."""
        x, y = mock_xy
        return gpd.GeoDataFrame(
            geometry=[box(x - 1, y - 1, x + 1, y + 1)], crs="EPSG:4326"
        )

    def test_parse_region(
        self,
        mock_get_basin_geometry,
        mock_datacatalog,
        mock_xy,
        mock_region,
    ):
        """Test the parsing of a basin region."""
        x, y = mock_xy

        # Act
        geoms, xy, ds_org = parse_region(
            data_catalog=mock_datacatalog,
            region=mock_region.copy(),
            hydrography_fn="mock_hydrography",
            # datacatalog is mocked so no need for a real file
        )

        # Assert
        for name, geom in geoms.items():
            assert isinstance(geom, gpd.GeoDataFrame), (
                "Parsed geometry should be a GeoDataFrame."
            )
            assert geom.crs.to_string() == "EPSG:4326", (
                "Geometry should have CRS EPSG:4326."
            )
            assert not geom.empty, "Returned geometry should not be empty."
            bounds = geom.total_bounds
            assert bounds[0] <= x <= bounds[2], (
                "X coordinate should fall within geometry bounds."
            )
            assert bounds[1] <= y <= bounds[3], (
                "Y coordinate should fall within geometry bounds."
            )

        assert isinstance(xy, np.ndarray), "XY should be a NumPy array."
        assert xy.shape == (1, 2), "XY should be a single (x, y) pair."
        assert np.allclose(xy[0], [x, y]), f"XY values should match input ({x}, {y})."

        assert isinstance(ds_org, xr.Dataset), (
            "Returned object should be an xarray Dataset."
        )
        assert "mask" in ds_org.coords, (
            "'mask' should be present in dataset coordinates."
        )
        mask = ds_org.coords["mask"].values
        assert mask.shape == (2, 2), "Mask should have expected shape (2, 2)."
        assert np.array_equal(mask, np.array([[1, 0], [0, 1]])), (
            "Mask values should match mocked output."
        )

    def test_parse_regions_different_kinds(
        self,
        mock_geometry,
        mock_region,
        mock_xy,
        mock_datacatalog,
        caplog,
        mocker,
    ):
        """Test parsing regions of different kinds."""
        caplog.set_level(logging.WARNING)
        x, y = mock_xy
        bbox_region = {"bbox": [x - 1, y - 1, x + 1, y + 1]}

        mock_parse_region_bbox = mocker.patch(
            "hydromt_wflow.workflows.basemaps.parse_region_bbox"
        )
        mock_parse_region_bbox.return_value = mock_geometry

        parsed_region = parse_region(
            data_catalog=mock_datacatalog,
            region=bbox_region,
            hydrography_fn="test",
        )

        assert (
            "Kind 'bbox' for the region is not recommended as it can lead "
            "to mistakes in the catchment delineation. Use carefully."
        ) in caplog.text

        assert isinstance(parsed_region[0], dict)
        assert parsed_region[1] is None

        geom_region = {"geom": mock_geometry.geometry}
        mock_parse_region_geom = mocker.patch(
            "hydromt_wflow.workflows.basemaps.parse_region_geom"
        )
        mock_parse_region_geom.return_value = mock_geometry

        parsed_region = parse_region(
            mock_datacatalog, region=geom_region, hydrography_fn="test"
        )

        assert (
            "Kind 'geom' for the region is not recommended as it can lead "
            "to mistakes in the catchment delineation. Use carefully."
        ) in caplog.text
        assert isinstance(parsed_region[0], dict)
        assert parsed_region[1] is None

    def test_parse_region_invalid_region_kind(self, mock_datacatalog, mock_region):
        """Test that an error is raised if the region kind is not supported."""
        interbasin_region = {"interbasin": mock_region.copy()["basin"]}
        err_msg = (
            "wflow region kind not understood or supported: interbasin. "
            "Use 'basin', 'subbasin', 'bbox' or 'geom'."
        )
        mock_raster_ds = mock_datacatalog.get_rasterdataset.return_value
        mock_raster_ds.raster.crs.is_geographic = False  # reset
        with pytest.raises(ValueError, match=err_msg):
            parse_region(
                mock_datacatalog,
                interbasin_region,
                hydrography_fn="hydrography_fn",
            )

    def test_parse_region_geometry_without_crs(
        self, mock_datacatalog, mock_get_basin_geometry, mock_region
    ):
        """Test that an error is raised if the basin geometry has no CRS."""
        basin_geom, xy = mock_get_basin_geometry.return_value
        basin_geom_no_crs = basin_geom.set_crs(None, allow_override=True)
        mock_get_basin_geometry.return_value = (basin_geom_no_crs, xy)

        with pytest.raises(ValueError, match="wflow region geometry has no CRS"):
            parse_region(
                mock_datacatalog,
                mock_region.copy(),
                hydrography_fn="hydrography_fn",
            )

    @pytest.mark.parametrize(
        ("resolution", "is_geographic", "expected_err_msg"),
        [
            (
                1 / 240,
                False,
                "Model resolution 0.004166666666666667 should be larger than the "
                "hydrography_fn resolution 0.008333333333333333.",
            ),  # Case 1: scale ratio << 1
            (
                2,
                True,
                r"The model resolution 2 should be smaller than 1 degree \(111km\) for "
                "geographic coordinate systems. Make sure you provided res in degree rather"  # noqa: E501
                " than in meters.",
            ),  # Case 2: scale ratio >> 1 (geographic crs)
        ],
    )
    def test_parse_region_scale_ratio_errors(
        self,
        mock_datacatalog,
        mock_region,
        resolution,
        is_geographic,
        expected_err_msg,
    ):
        """Test that an error is raised if the model resolution is not compatible with the hydrography resolution."""  # noqa: E501
        mock_raster_ds = mock_datacatalog.get_rasterdataset.return_value
        mock_raster_ds.raster.crs.is_geographic = is_geographic
        with pytest.raises(ValueError, match=expected_err_msg):
            parse_region(
                mock_datacatalog,
                region=mock_region.copy(),
                hydrography_fn="hydrography_fn",
                resolution=resolution,
            )

    def test_parse_region_scale_ratio_close_to_one_no_error(
        self,
        mock_datacatalog,
        mock_get_basin_geometry,
        mock_xy,
        mock_region,
        mock_geometry,
    ):
        """Test that no error is raised if the model resolution is close to the hydrography resolution."""  # noqa: E501
        mock_geometry.set_crs(crs="EPSG:4326")  # restore valid CRS
        mock_get_basin_geometry.return_value = (mock_geometry, mock_xy)
        resolution = mock_datacatalog.get_rasterdataset.return_value.raster.res[0]
        # Should not raise any errors or warnings
        parse_region(
            mock_datacatalog,
            mock_region.copy(),
            hydrography_fn="test",
            resolution=resolution,
        )

    def test_parse_region_scale_ratio_warning(
        self,
        mock_get_basin_geometry,
        mock_datacatalog,
        mock_region,
        caplog,
    ):
        """Test that a warning is logged if the model resolution is slightly smaller than the hydrography resolution."""  # noqa: E501
        hydrography_resolution = (
            mock_datacatalog.get_rasterdataset.return_value.raster.res[0]
        )
        slightly_smaller_resolution = hydrography_resolution * 0.9  # scale_ratio ~ 1.11

        with caplog.at_level(logging.WARNING):
            parse_region(
                mock_datacatalog,
                mock_region.copy(),
                hydrography_fn="test",
                resolution=slightly_smaller_resolution,
            )
        assert any(
            "Model resolution" in rec.message
            and "does not match the hydrography resolution" in rec.message
            for rec in caplog.records
        )



import logging
from collections.abc import Callable
from pathlib import Path

import pytest
import xarray as xr

from hydromt_wflow.components import WflowOutputCsvComponent
from hydromt_wflow.wflow_sbm import WflowSbmModel


@pytest.fixture
def mock_model(mock_model_factory) -> WflowSbmModel:
    """Fixture to create a mock WflowSbmModel."""
    return mock_model_factory(mode="w")


# Define the side effect function
def get_value_side_effect(key, fallback=None):
    if key == "dir_output":
        return "run_results"
    elif key == "output.csv.path":
        return "output.csv"
    elif key == "dir_input":
        return ""
    return fallback


def test_wflow_output_csv_component_init(mock_model: WflowSbmModel):
    # Setup the component
    component = WflowOutputCsvComponent(mock_model)

    # Assert that the internal data is None
    assert component._data is None

    # When asking for data property, it should return a dict
    assert isinstance(component.data, dict)
    assert isinstance(component._data, dict)  # Same for internal
    assert len(component.data) == 0


def test_wflow_output_csv_component_init_with_locs(
    mock_model_staticmaps: WflowSbmModel,
):
    # Setup the component with a region component
    component = WflowOutputCsvComponent(
        mock_model_staticmaps, locations_component="staticmaps"
    )

    # Check if the locations component is set correctly
    assert component._locations_component == "staticmaps"
    assert "basin" in component._get_locations_data()


def test_wflow_output_csv_component_set(
    mock_model_staticmaps: WflowSbmModel, forcing_layer: xr.DataArray
):
    # Setup the component
    component = WflowOutputCsvComponent(
        mock_model_staticmaps, locations_component="staticmaps"
    )

    # Assert that the internal data is None, initializing will happen in set
    assert component._data is None

    # Set an entry
    # Copy else the name gets altered for the next test?
    component.set(forcing_layer.copy(), name="test_layer")

    # Assert the content
    assert isinstance(component._data, dict)
    assert "test_layer" in component.data
    assert len(component.data) == 1
    assert isinstance(component.data["test_layer"], xr.DataArray)


def test_wflow_output_csv_component_read(
    mock_model_staticmaps_config_factory: Callable[[Path, str], WflowSbmModel],
    model_subbasin_cached: Path,
):
    # Set it to read mode
    mock_model = mock_model_staticmaps_config_factory(
        path=model_subbasin_cached, mode="r", config_filename="wflow_sbm_results.toml"
    )
    # Here we really need the config and the staticmaps to read...
    assert "dir_output" in mock_model.config.data
    assert "subcatchment" in mock_model.staticmaps.data

    # Setup the component
    component = WflowOutputCsvComponent(mock_model, locations_component="staticmaps")

    # Assert its data currently none
    assert component._data is None

    # Read the data
    component.read()

    # Assert the read data
    assert isinstance(component.data, dict)
    assert len(component.data) == 6
    assert "river_max_q" in component.data


def test_wflow_output_csv_component_read_logging(
    mock_model_staticmaps_config_factory: Callable[[Path, str], WflowSbmModel],
    model_subbasin_cached: Path,
    caplog: pytest.LogCaptureFixture,
):
    # Set it to read mode
    mock_model = mock_model_staticmaps_config_factory(
        path=model_subbasin_cached, mode="r", config_filename="wflow_sbm_results.toml"
    )

    # Setup the component
    component = WflowOutputCsvComponent(mock_model, locations_component="staticmaps")

    # Missing map in staticmaps
    mock_model.staticmaps.drop_vars("gauges_grdc")

    with caplog.at_level(logging.DEBUG, logger="hydromt.hydromt_wflow.utils"):
        # Read the data
        component.read()

    assert "Reading csv column 'river_max_q'" in caplog.text
    assert "Map 'gauges_grdc' not found in staticmaps. Skip reading." in caplog.text
    assert len(component.data) == 0

    # Empty staticmaps
    mock_model.staticmaps._data = xr.Dataset()
    with caplog.at_level(logging.WARNING):
        component.read()

    assert "Staticmaps data is empty, skip reading." in caplog.text
    assert len(component.data) == 0


def test_wflow_output_csv_component_write(
    mock_model: WflowSbmModel,
    grid_dummy_data: xr.DataArray,
    caplog: pytest.LogCaptureFixture,
):
    # Setup the component
    component = WflowOutputCsvComponent(mock_model)
    component._data = {
        "test_layer": grid_dummy_data,
    }

    # Write to a file
    # Check the logger message when calling component.write()
    with caplog.at_level(logging.INFO):
        component.write()
    assert "output_csv is an output of Wflow and will not be written." in caplog.text



from hydromt import DataCatalog


def test_fetch_data(build_data_catalog, build_data):
    assert build_data_catalog.is_file()
    assert build_data.is_dir()
    dc = DataCatalog(build_data_catalog)
    assert dc.get_source("chelsa").name == "chelsa"
